from torch.nn import MultiheadAttention
from torch.nn.modules import Module
from zennit.attribution import Attributor
from zennit.composites import LayerMapComposite

from pnpxai.explainers.attentions.rules import SavingAttention
from pnpxai.explainers.attentions.module_converters import default_attention_converters
from pnpxai.explainers.utils import _format_to_tuple


class SavingAttentionAttributor(Attributor):
    def __init__(self, model: Module):
        layer_map = [(MultiheadAttention, SavingAttention())]
        composite = LayerMapComposite(layer_map=layer_map, canonizers=default_attention_converters)
        super().__init__(model, composite, None)    

    
    def forward(self, input, attr_output_fn):
        input = _format_to_tuple(input)
        _ = self.model(*input)
        attn_output_weights_all = [
            hook_ref.stored_tensors[hook_ref.saved_name]
            for hook_ref in self.composite.hook_refs
        ]
        return tuple(attn_output_weights_all)



from typing import Any, Optional, Dict

import torch
from torch import Tensor, nn

from pnpxai.core.detector.types import Linear, Convolution
from pnpxai.core._types import Model, DataSource, Task
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.rap.rap import RelativeAttributePropagation


class RAP(Explainer):
    """
    Computes Relative Attribute Propagation (RAP) explanations for a given model.

    Supported Modules: `Linear`, `Convolution`

    Parameters:
        model (Model): The model for which RAP explanations are computed.

    Reference:
        Woo-Jeoung Nam, Shir Gur, Jaesik Choi, Lior Wolf, Seong-Whan Lee. Relative Attributing Propagation: Interpreting the Comparative Contributions of Individual Units in Deep Neural Networks.
    """

    SUPPORTED_MODULES = [Linear, Convolution]

    def __init__(self, model: Model):
        super().__init__(model)
        self.method = RelativeAttributePropagation(model)

    def compute_pred(self, output: Tensor) -> Tensor:
        """
        Computes the predicted class probabilities.

        Parameters:
            output (Tensor): The model output.

        Returns:
            Tensor: The one-hot encoded predicted class probabilities.
        """
        # get the index of the max log-probability
        pred = output.max(1, keepdim=True)[1]
        pred = pred.squeeze(-1)

        pred_one_hot = nn.functional.one_hot(pred, output.shape[-1]) * 1.0
        pred_one_hot = pred_one_hot.to(output.device)
        return pred_one_hot

    def attribute(self, inputs: DataSource, targets: DataSource, *args: Any, **kwargs: Any) -> DataSource:
        """
        Computes RAP attributions for the given inputs.

        Parameters:
            inputs (DataSource): The input data.
            targets (DataSource): The target labels.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            DataSource: RAP attributions.
        """
        outputs = self.method.run(inputs)
        preds = self.compute_pred(outputs)
        relprop = self.method.relprop(preds)
        return relprop

    def format_outputs_for_visualization(
        self,
        inputs: DataSource,
        targets: DataSource,
        explanations: DataSource,
        task: Task,
        kwargs: Optional[Dict[str, Any]] = None,
    ):
        explanations = explanations.detach().cpu()\
            .transpose(-1, -3)\
            .transpose(-2, -3)

        explanations = explanations.sum(-1)
        agg_dims = list(range(1, explanations.ndim))
        pos_max = explanations.relu().amax(agg_dims, keepdim=True)
        neg_max = (-explanations).relu().amax(agg_dims, keepdim=True)

        explanations = torch.where(
            explanations > 0,
            explanations / pos_max,
            explanations / neg_max
        )

        return explanations.numpy()



from typing import (
	Optional,
	Callable,
	Tuple,
	Union,
	Sequence,
)

from torch import Tensor
from torch.nn.modules import Module

from pnpxai.core.detector.types import Linear, Convolution, LSTM, RNN, Attention
from pnpxai.utils import format_into_tuple, format_out_tuple_if_single
from pnpxai.explainers.smooth_grad import SmoothGrad


class VarGrad(SmoothGrad):
	"""
    VarGrad explainer.

    Supported Modules: `Linear`, `Convolution`, `LSTM`, `RNN`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
		noise_level (float): The noise level added during attribution.
		n_iter (int): The number of iterations, the input is modified.
		layer (Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]): The target module to be explained.
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).		
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Lorenz Richter, Ayman Boustati, Nikolas Nüsken, Francisco J. R. Ruiz, Ömer Deniz Akyildiz. VarGrad: A Low-Variance Gradient Estimator for Variational Inference.
    """

	SUPPORTED_MODULES = [Linear, Convolution, LSTM, RNN, Attention]

	def __init__(
		self,
		model: Module,
		noise_level: float=.1,
		n_iter: int=20,
        forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]]=None,
        additional_forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]]=None,
        layer: Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]=None,
        n_classes: Optional[int]=None,
	) -> None:
		super().__init__(
			model=model,
			noise_level=noise_level,
			n_iter=n_iter,
			forward_arg_extractor=forward_arg_extractor,
			additional_forward_arg_extractor=additional_forward_arg_extractor,
			layer=layer,
			n_classes=n_classes,
		)

	def attribute(
		self,
		inputs: Union[Tensor, Tuple[Tensor]],
		targets: Tensor,
	) -> Union[Tensor, Tuple[Tensor]]:
		"""
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            torch.Tensor: The result of the explanation.
        """
		forward_args, additional_forward_args = self._extract_forward_args(inputs)
		with self.attributor() as attributor:
			avg_grads, avg_grads_sq = attributor.forward(
				forward_args,
				targets,
				additional_forward_args,
				return_squared=True,
			)
		vargrads = tuple(
			avg_grad_sq - avg_grad for avg_grad_sq, avg_grad in zip(
				format_into_tuple(avg_grads_sq),
				format_into_tuple(avg_grads),
			)
		)
		return format_out_tuple_if_single(vargrads)



from importlib import util
from pnpxai.explainers.zennit.module_converter import module_converting_canonizer_factory

cfgs = dict()

if util.find_spec("timm"):
    import timm
    from pnpxai.explainers.attentions.module_converters.configs import TIMM_VIT_ATTENTION_CONVERTER_FACTORY_CONFIG
    cfgs[timm.models.vision_transformer.Attention] = TIMM_VIT_ATTENTION_CONVERTER_FACTORY_CONFIG

if util.find_spec("transformers"):
    import transformers
    from pnpxai.explainers.attentions.module_converters.configs import (
        TRANSFOMERS_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG,
        TRANSFORMERS_VISUAL_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG,
        TRANSFORMERS_VILT_LAYER_CONVERTER_FACTORY_CONFIG,
        TRANSFORMERS_LXMERT_SELF_ATTENTION_CONVERTER_FACTORY_CONFIG,
        TRANSFORMERS_LXMERT_CROSS_ATTENTION_CONVERTER_FACTORY_CONFIG,
    )
    cfgs[transformers.models.bert.modeling_bert.BertAttention] = TRANSFOMERS_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG
    cfgs[transformers.models.visual_bert.modeling_visual_bert.VisualBertAttention] = TRANSFORMERS_VISUAL_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG
    cfgs[transformers.models.vilt.modeling_vilt.ViltLayer] = TRANSFORMERS_VILT_LAYER_CONVERTER_FACTORY_CONFIG
    cfgs[transformers.models.lxmert.modeling_lxmert.LxmertSelfAttentionLayer] = TRANSFORMERS_LXMERT_SELF_ATTENTION_CONVERTER_FACTORY_CONFIG
    cfgs[transformers.models.lxmert.modeling_lxmert.LxmertCrossAttentionLayer] = TRANSFORMERS_LXMERT_CROSS_ATTENTION_CONVERTER_FACTORY_CONFIG

# import pdb; pdb.set_trace()

default_attention_converters = [
    module_converting_canonizer_factory(**cfg)()
    for cfg in cfgs.values()
] or None


from typing import Optional
from pnpxai.core.experiment.manager import ExperimentManager
from pnpxai.explainers.base import Explainer
from pnpxai.evaluator.metrics.base import Metric


class ExperimentObservableEvent:
    def __init__(
        self,
        manager: ExperimentManager,
        message: str,
        explainer: Optional[Explainer] = None,
        metric: Optional[Metric] = None
    ):
        self.message = message
        self.progress = self._compute_progress(manager, explainer, metric)

    def _compute_progress(
        self,
        manager: ExperimentManager,
        explainer: Optional[Explainer] = None,
        metric: Optional[Metric] = None
    ):
        explainers = manager.get_explainers()[0]
        metrics = manager.get_metrics()[0]

        progress = 0
        # Explainer has len(metrics) computations and its own
        n_explainer_states = len(metrics) + 1
        total_len = len(explainers) * (n_explainer_states)
        if explainer is not None:
            explainer_idx = explainers.index(explainer)
            # Offset number of computations before current and add current computation
            progress += explainer_idx * n_explainer_states + 1

        if metric is not None:
            metric_idx = metrics.index(metric)
            progress += metric_idx + 1

        return progress / total_len



from abc import abstractmethod

import torch
from torch.nn import MultiheadAttention
from zennit.core import Stabilizer
from zennit.rules import ClampMod, zero_bias, ParamMod
from zennit.core import RemovableHandleList, RemovableHandle, Hook

from pnpxai.explainers.attentions import partial_forwards
from pnpxai.explainers.zennit.hooks import HookWithKwargs


class SavingAttention(HookWithKwargs):
    def __init__(self, saved_name: str="attn_output_weights"):
        super().__init__()
        self.saved_name = saved_name

    def pre_forward(self, module, input, kwargs):
        kwargs["need_weights"] = True
        kwargs["average_attn_weights"] = False
        return super().pre_forward(module, input, kwargs)

    def forward(self, module, input, kwargs, output):
        self.stored_tensors[self.saved_name] = output[-1]

    def register(self, module):
        '''Register this instance by registering all hooks to the supplied module.'''
        return RemovableHandleList([
            RemovableHandle(self),
            module.register_forward_pre_hook(self.pre_forward, with_kwargs=True),
            module.register_forward_hook(self.post_forward),
            module.register_forward_hook(self.forward, with_kwargs=True),
        ])


class AttentionRuleBase(HookWithKwargs):
    def __init__(
            self,
            stabilizer=1e-6,
            **kwargs
        ):
        super().__init__()
        self.stabilizer = stabilizer
        self.stabilizer_fn = Stabilizer.ensure(stabilizer)

    @abstractmethod
    def backward(self, module: MultiheadAttention, grad_input, grad_output):
        raise NotImplementedError



class ConservativeAttentionPropagation(AttentionRuleBase):
    def __init__(self, stabilizer=1e-6):
        super().__init__(stabilizer)

    def backward(self, module, grad_input, grad_output):
        query, key, value = (
            inp.clone().requires_grad_()
            for inp in self.stored_tensors["input"]
        )
        with torch.autograd.enable_grad():
            preprocessed_inputs = partial_forwards.preprocess_inputs(
                module,
                query,
                key,
                value,
                **self.stored_kwargs
            )
            query, key, value, key_padding_mask, attn_mask, is_batched = preprocessed_inputs
            tgt_len, bsz, _ = query.size()
            src_len, _, _ = key.size()

            # forward
            in_proj_query, in_proj_key, in_proj_value = partial_forwards.in_proj_qkv(
                module, query, key, value,
            )
            attn_output_weights = partial_forwards.attn_output_weights(
                module, in_proj_query, in_proj_key,
                module.bias_k, key_padding_mask, attn_mask,
                tgt_len, bsz, src_len,
            )
            attn_output_weights = attn_output_weights.detach()
            in_proj_output = partial_forwards.in_proj_output(
                module, in_proj_value, attn_output_weights,
                tgt_len, bsz,
            )
            out_proj = partial_forwards.out_proj(
                module, in_proj_output, tgt_len, bsz, is_batched
            )

        # relprop: out_proj -> in_proj_output
        grad_outputs = grad_output[0] / self.stabilizer_fn(out_proj)
        gradients = torch.autograd.grad(
            out_proj,
            in_proj_output,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        rel_in_proj_output = in_proj_output * gradients[0]

        # relprop: in_proj_output -> in_proj_value, attn_output_weights
        grad_outputs = rel_in_proj_output / self.stabilizer_fn(in_proj_output)
        gradients = torch.autograd.grad(
            in_proj_output,
            in_proj_value,
            grad_outputs=grad_outputs,
        )
        rel_in_proj_value = in_proj_value * gradients[0]

        '''
        rel_attn_output_weights = attn_output_weights * gradients[1]

        # relprop: attn_output_weights -> in_proj_query, in_proj_key
        grad_outputs = rel_attn_output_weights / self.stabilizer_fn(attn_output_weights)
        gradients = torch.autograd.grad(
            attn_output_weights,
            (in_proj_query, in_proj_key),
            grad_outputs=grad_outputs,
        )
        rel_in_proj_query = in_proj_query * gradients[0]
        rel_in_proj_key = in_proj_key * gradients[1]

        # relprop: in_proj_query -> query
        grad_outputs = rel_in_proj_query / self.stabilizer_fn(in_proj_query)
        gradients = torch.autograd.grad(
            in_proj_query,
            query,
            grad_outputs=grad_outputs,
        )
        rel_query = query * gradients[0]

        # relprop: in_proj_key -> key
        grad_outputs = rel_in_proj_key / self.stabilizer_fn(in_proj_key)
        gradients = torch.autograd.grad(
            in_proj_key,
            key,
            grad_outputs=grad_outputs,
        )
        rel_key = key * gradients[0]
        '''

        # relprop: in_proj_value -> value
        grad_outputs = rel_in_proj_value / self.stabilizer_fn(in_proj_value)
        gradients = torch.autograd.grad(
            in_proj_value,
            value,
            grad_outputs=grad_outputs,
        )

        rel_value = value * gradients[0]

        if module.batch_first and is_batched:
            query, key, rel_value = (
                x.transpose(1, 0) for x in (query, key, rel_value)
            )
        return (
            torch.zeros_like(query),
            rel_value,
            rel_value,
        )

    def copy(self):
        copy = ConservativeAttentionPropagation.__new__(type(self))
        ConservativeAttentionPropagation.__init__(
            copy,
            self.stabilizer
        )
        return copy


class LayerNormRule(Hook):
    def __init__(self, stabilizer=1e-6):
        super().__init__()
        self.stabilizer = stabilizer

    def forward(self, module, input, output):
        '''Forward hook to save module in-/outputs.'''
        self.stored_tensors['input'] = input

    def backward(self, module, grad_input, grad_output):
        input = self.stored_tensors["input"][0].clone().requires_grad_()
        with torch.autograd.enable_grad():
            # Preprocess
            dims = tuple(range(-len(module.normalized_shape), -2, -1))
            
            # Normalization: centering
            mean = input.mean(dims, keepdims=True)
            centered = input - mean

            # Normalization: rescaling
            # correction should be 1 as it is calculated via the unbiased estimator.
            var = input.var(dims, keepdims=True, correction=1).detach()
            rescaled = centered / (var + module.eps).sqrt()

            # Elementwise affine transformation
            transformed = torch.mul(rescaled, module.weight) + module.bias

        stabilizer_fn = Stabilizer.ensure(self.stabilizer)

        ''' # Separate relevance propagation including affine transformation
        # Relevance: rescaled
        grad_outputs = grad_output[0] / stabilizer_fn(transformed)
        gradients = torch.autograd.grad(
            transformed,
            rescaled,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_rescaled = rescaled * gradients[0]

        # Relevance: input
        # pdb.set_trace()
        grad_outputs = relevance_rescaled[0] / stabilizer_fn(rescaled)
        gradients = torch.autograd.grad(
            rescaled,
            input,
            grad_outputs=grad_outputs,
        )
        relevance_input = input * gradients[0]
        '''

        # ''' # End-to-end relevance propagation including affine transformation
        # Relevance: end-to-end
        grad_outputs = grad_output[0] / stabilizer_fn(transformed)
        gradients = torch.autograd.grad(
            transformed,
            input,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_input = input * gradients[0]
        # '''

        ''' # Relevance: end-to-end without affine transformation
        grad_outputs = grad_output[0] / stabilizer_fn(rescaled)
        gradients = torch.autograd.grad(
            rescaled,
            input,
            grad_outputs=grad_outputs,
            create_graph=grad_output[0].requires_grad,
        )
        relevance_input = input * gradients[0]
        '''
        return tuple(relevance_input if original.shape == relevance_input.shape else None for original in grad_input)


class CGWAttentionPropagation(AttentionRuleBase):
    '''
    ``zennit``-compatible attention propagation rule of Chefer, Gur and Wolf.
    '''
    def __init__(
            self,
            alpha=2.,
            beta=1.,
            stabilizer=1e-6,
            save_attn_output_weights=False,
            zero_params=None,
        ):
        super().__init__(stabilizer)
        self.alpha = alpha
        self.beta = beta
        self.stabilizer = stabilizer
        self.save_attn_output_weights = save_attn_output_weights
        self.input_clampers = [
            lambda input: input.clamp(min=0), # px
            lambda input: input.clamp(max=0), # nx
            lambda input: input.clamp(min=0), # px
            lambda input: input.clamp(max=0), # nx
        ]
        self.param_clampers = [
            ClampMod(min=0., zero_params=zero_params), # pw
            ClampMod(max=0., zero_params=zero_bias(zero_params)), # nw
            ClampMod(max=0., zero_params=zero_bias(zero_params)), # nw
            ClampMod(min=0., zero_params=zero_params), # pw
        ]

    def backward(self, module, grad_input, grad_output):
        query, key, value = (
            inp.clone().requires_grad_()
            for inp in self.stored_tensors["input"]
        )
        with torch.autograd.enable_grad():
            preprocessed_inputs = partial_forwards.preprocess_inputs(
                module,
                query,
                key,
                value,
                **self.stored_kwargs
            )
            query, key, value, key_padding_mask, attn_mask, is_batched = preprocessed_inputs
            tgt_len, bsz, _ = query.size()
            src_len, _, _ = key.size()

            # forward
            in_proj_query, in_proj_key, in_proj_value = partial_forwards.in_proj_qkv(
                module, query, key, value,
            )
            attn_output_weights = partial_forwards.attn_output_weights(
                module, in_proj_query, in_proj_key,
                module.bias_k, key_padding_mask, attn_mask,
                tgt_len, bsz, src_len,
            )
            in_proj_output = partial_forwards.in_proj_output(
                module, in_proj_value, attn_output_weights,
                tgt_len, bsz,
            )
            out_proj = partial_forwards.out_proj(
                module, in_proj_output, tgt_len, bsz, is_batched
            )

            # clamped in_proj_qkv
            clamped_qkvs = []
            clamped_in_proj_qkvs = []
            for input_clamper, param_clamper in zip(self.input_clampers, self.param_clampers):
                clamped_qkv = (input_clamper(query), input_clamper(key), input_clamper(value))
                with ParamMod.ensure(param_clamper)(module) as clamped_module:
                    clamped_in_proj_qkv = partial_forwards.in_proj_qkv(clamped_module, *clamped_qkv)
                clamped_qkvs.append(clamped_qkv)
                clamped_in_proj_qkvs.append(clamped_in_proj_qkv)

            # clamped out_proj
            clamped_in_proj_outputs = []
            clamped_out_projs = []
            for input_clamper, param_clamper in zip(self.input_clampers, self.param_clampers):
                clamped_in_proj_output = input_clamper(in_proj_output)
                with ParamMod.ensure(param_clamper)(module.out_proj) as clamped_out_proj_forward:
                    clamped_out_proj = clamped_out_proj_forward(clamped_in_proj_output)
                clamped_in_proj_outputs.append(clamped_in_proj_output)
                clamped_out_projs.append(clamped_out_proj.view(tgt_len, bsz, -1).transpose(0, 1))

        # save attention output weights
        if self.save_attn_output_weights:
            self.stored_tensors["attn_output_weights"] = attn_output_weights.view(bsz, module.num_heads, tgt_len, -1)

        # save attention gradients
        attn_gradients = torch.autograd.grad(
            out_proj,
            attn_output_weights,
            grad_outputs=torch.ones_like(out_proj),
            create_graph=grad_output[0].requires_grad,
            retain_graph=True,
        )
        self.stored_tensors["attn_grads"] = attn_gradients[0].view(bsz, module.num_heads, tgt_len, -1)

        # relevance propagation: out_proj - in_proj_output / alpha-beta rule
        # activator
        act_rel = torch.zeros_like(clamped_in_proj_outputs[0])
        for clamped_in_proj_output, clamped_out_proj in zip(clamped_in_proj_outputs[:2], clamped_out_projs[:2]):
            grad_outputs = grad_output[0] / self.stabilizer_fn(clamped_out_proj)
            gradients = torch.autograd.grad(
                clamped_out_proj,
                clamped_in_proj_output,
                grad_outputs=grad_outputs,
                create_graph=grad_output[0].requires_grad,
            )
            act_rel += clamped_in_proj_output * gradients[0]

        # inhibitor
        inhb_rel = torch.zeros_like(clamped_in_proj_outputs[0])
        for clamped_in_proj_output, clamped_out_proj in zip(clamped_in_proj_outputs[2:], clamped_out_projs[2:]):
            grad_outputs = grad_output[0] / self.stabilizer_fn(clamped_out_proj)
            gradients = torch.autograd.grad(
                clamped_out_proj,
                clamped_in_proj_output,
                grad_outputs=grad_outputs,
                create_graph=grad_output[0].requires_grad,
            )
            inhb_rel += clamped_in_proj_output * gradients[0]
        rel_in_proj_output = self.alpha * act_rel - self.beta * inhb_rel

        # relevance propagation: in_proj_value, attn_output_weights / zero rule
        grad_outputs = rel_in_proj_output / self.stabilizer_fn(in_proj_output)
        gradients = torch.autograd.grad(
            in_proj_output,
            (in_proj_value, attn_output_weights),
            grad_outputs=grad_outputs,
        )
        rel_in_proj_value = in_proj_value * gradients[0]
        rel_attn_output_weights = attn_output_weights * gradients[1]

        rel_in_proj_value /= 2
        rel_attn_output_weights /= 2

        self.stored_tensors["attn_rels"] = rel_attn_output_weights.view(bsz, module.num_heads, tgt_len, -1)

        # relevance propagation: in_proj_query, in_proj_value / zero rule
        grad_outputs = rel_attn_output_weights / self.stabilizer_fn(attn_output_weights)
        gradients = torch.autograd.grad(
            attn_output_weights,
            (in_proj_query, in_proj_key),
            grad_outputs=grad_outputs,
        )
        rel_in_proj_query = in_proj_query * gradients[0]
        rel_in_proj_key = in_proj_key * gradients[1]

        rel_in_proj_query /= 2
        rel_in_proj_key /= 2

        rel_in_proj_qkvs = (rel_in_proj_query, rel_in_proj_key, rel_in_proj_value)

        # relevance propagation: in_proj_qkv / alpha-beta rule
        rel_in_proj_qkv_iterator = zip(
            zip(*clamped_qkvs),
            zip(*clamped_in_proj_qkvs),
            rel_in_proj_qkvs,
        )
        rel_qkvs = []
        for clamped_ins, clamped_outs, rel_in_proj in rel_in_proj_qkv_iterator:
            '''
            The iterator sequentially propagates relevance following alpha-beta rule
                clamped_ins: [p, n, p, n] of {query, key, value}
                clamped_outs: [pp, nn, pn, np] of {query, key, value}
                rel_in_proj: relevance scores of {query, key, value} propagated right before
            '''
            # activator
            act_rel = torch.zeros_like(clamped_ins[0])
            for clamped_in, clamped_out in zip(clamped_ins[:2], clamped_outs[:2]):
                grad_outputs = rel_in_proj / self.stabilizer_fn(clamped_out)
                gradients = torch.autograd.grad(
                    clamped_out,
                    clamped_in,
                    grad_outputs=grad_outputs,
                )
                act_rel += clamped_in * gradients[0]

            # inhibitor
            inhb_rel = torch.zeros_like(clamped_ins[0])
            for clamped_in, clamped_out in zip(clamped_ins[2:], clamped_outs[2:]):
                grad_outputs = rel_in_proj / self.stabilizer_fn(clamped_out)
                gradients = torch.autograd.grad(
                    clamped_out,
                    clamped_in,
                    grad_outputs=grad_outputs,
                )
                inhb_rel += clamped_in * gradients[0]
            rel = self.alpha * act_rel - self.beta * inhb_rel
            rel_qkvs.append(rel)

        if module.batch_first and is_batched:
            rel_qkvs = tuple(rel.transpose(0, 1) for rel in rel_qkvs)
        return rel_qkvs


    def copy(self):
        '''Return a copy of this hook.
        This is used to describe hooks of different modules by a single hook instance.
        '''
        copy = CGWAttentionPropagation.__new__(type(self))
        CGWAttentionPropagation.__init__(
            copy,
            self.alpha,
            self.beta,
            self.stabilizer,
            self.save_attn_output_weights,
        )
        return copy


import torch
import timm
import transformers
from torch import nn

from pnpxai.explainers.attentions.module_converters.layers import StackAndSum
from pnpxai.explainers.attentions.module_converters.validations import *
from pnpxai.explainers.attentions.module_converters.warnings import *


def transform_attn_mask_of_transformers(in_args, in_kwargs, kept, out_module):
    # required: (tgt_len, src_len) or (bsz*num_heads, tgt_len, src_len)
    attn_mask = in_kwargs.get("attention_mask") or in_args[1]
    if attn_mask is None:
        return
    bsz, num_heads, tgt_len, src_len = attn_mask.shape
    if bsz == 1 and num_heads == 1 and tgt_len == 1: # self attention
        attn_mask = attn_mask.repeat(1, 1, src_len, 1)
        return attn_mask.squeeze()
    elif bsz > 1 and num_heads == 1 and tgt_len == 1:
        attn_mask = attn_mask.repeat(
            1, out_module._converted_self_attention.num_heads,
            src_len, 1
        )
        return attn_mask.view(-1, src_len, src_len)
    elif bsz > 1 and num_heads == 1 and tgt_len > 1:
        attn_mask = attn_mask.repeat(
            1, out_module._converted_self_attention.num_heads,
            1, 1,
        )
        return attn_mask.view(-1, tgt_len, src_len)
    else:
        raise Exception("Not covered yet")


# timm vit
TIMM_VIT_ATTENTION = {
    "out_module_type": nn.MultiheadAttention,
    "args": {
        "embed_dim": lambda in_module: in_module.head_dim * in_module.num_heads,
        "num_heads": lambda in_module: in_module.num_heads,
        "dropout": lambda in_module: in_module.attn_drop.p,
        "bias": lambda in_module: (in_module.qkv.bias.count_nonzero() > 0).item(),
        "add_bias_kv": False,
        "add_zero_attn": False,
        "kdim": None,
        "vdim": None,
        "batch_first": True,
        "device": lambda in_module: next(in_module.parameters()).device,
        "dtype": None,
    },
    "params": {
        "in_proj_weight": "qkv.weight",
        "out_proj.weight": "proj.weight",
        "in_proj_bias": "qkv.bias",
        "out_proj.bias": "proj.bias",
    },
    "forward": {
        "args": lambda in_args: tuple(in_args[0] for _ in range(3)),
        "kwargs": {
            "key_padding_mask": None,
            "need_weights": False,
            "attn_mask": None,
            "average_attn_weights": True,
            "is_causal": False,
        },
        "validations": [],
        "warnings": [],
    },
    "validations": [],
    "warnings": [],
}
TIMM_VIT_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": timm.models.vision_transformer.Attention,
    "config_selector": True,
    "out_module_configs": {
        True: {
            "_converted_self_attention": TIMM_VIT_ATTENTION,
        },
    },
}

# transformers bert
TRANSFORMERS_BERT_ATTENTION = {
    "out_modules": {
        "_converted_self_attention": {
            "out_module_type": nn.MultiheadAttention,
            "args": {
                "embed_dim":  lambda in_module: in_module.self.attention_head_size * in_module.self.num_attention_heads,
                "num_heads": lambda in_module: in_module.self.num_attention_heads,
                "dropout": lambda in_module: in_module.self.dropout.p,
                "bias": True,
                "add_bias_kv": False,
                "add_zero_attn": False,
                "kdim": None,
                "vdim": None,
                "batch_first": True,
                "device": lambda in_module: next(in_module.self.parameters()).device,
                "dtype": None,
            },
            "params": {
                "in_proj_weight": lambda in_module_params: torch.cat([
                    in_module_params["self.query.weight"],
                    in_module_params["self.key.weight"],
                    in_module_params["self.value.weight"],
                ]),
                "in_proj_bias": lambda in_module_params: torch.cat([
                    in_module_params["self.query.bias"],
                    in_module_params["self.key.bias"],
                    in_module_params["self.value.bias"]
                ]),
                "out_proj.weight": "output.dense.weight",
                "out_proj.bias": "output.dense.bias",
            },
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: tuple(
                    in_args[0] for _ in range(3)
                ),
                "kwargs": {
                    "key_padding_mask": None,
                    "need_weights": False,
                    "attn_mask": transform_attn_mask_of_transformers,
                    "average_attn_weights": True,
                    "is_causal": False,
                },
                "keep_outputs": True,
            },
        },
        "_converted_dropout": {
            "module": lambda in_module: in_module.output.dropout,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_self_attention"][0],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_layernorm": {
            "module": lambda in_module: in_module.output.LayerNorm,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_dropout"] + in_args[0],
                ),
                "keep_outputs": True,
            },
        },
    },
    "output_selector": lambda kept: (
        kept["outputs"]["_converted_layernorm"], # attention output
        kept["outputs"]["_converted_self_attention"][1], # attention weights
    ),
}
TRANSFOMERS_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.bert.modeling_bert.BertAttention,
    "config_selector": lambda in_module: True,
    "out_module_configs": {True: TRANSFORMERS_BERT_ATTENTION},
}

# transformers visual bert
TRANSFORMER_VISUAL_BERT_ATTENTION = TRANSFORMERS_BERT_ATTENTION
TRANSFORMERS_VISUAL_BERT_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.visual_bert.modeling_visual_bert.VisualBertAttention,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMER_VISUAL_BERT_ATTENTION,
    },
}

# transformers vilt
TRANSFORMERS_VILT_ATTENTION = {
    "out_modules": {
        "_converted_self_attention": {
            "out_module_type": nn.MultiheadAttention,
            "args": {
                "embed_dim":  lambda in_module: in_module.attention.attention_head_size * in_module.attention.num_attention_heads,
                "num_heads": lambda in_module: in_module.attention.num_attention_heads,
                "dropout": lambda in_module: in_module.attention.dropout.p,
                "bias": True,
                "add_bias_kv": False,
                "add_zero_attn": False,
                "kdim": None,
                "vdim": None,
                "batch_first": True,
                "device": lambda in_module: next(in_module.attention.parameters()).device,
                "dtype": None,
            },
            "params": {
                "in_proj_weight": lambda in_module_params: torch.cat([
                    in_module_params["attention.query.weight"],
                    in_module_params["attention.key.weight"],
                    in_module_params["attention.value.weight"],
                ]),
                "in_proj_bias": lambda in_module_params: torch.cat([
                    in_module_params["attention.query.bias"],
                    in_module_params["attention.key.bias"],
                    in_module_params["attention.value.bias"]
                ]),
                "out_proj.weight": "output.dense.weight",
                "out_proj.bias": "output.dense.bias",
            },
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: tuple(
                    in_args[0] for _ in range(3)
                ),
                "kwargs": {
                    "key_padding_mask": None,
                    "need_weights": False,
                    "attn_mask": transform_attn_mask_of_transformers,
                    "average_attn_weights": True,
                    "is_causal": False,
                },
                "keep_outputs": True,
            },
        },
        "_converted_dropout": {
            "module": lambda in_module: in_module.output.dropout,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (kept["outputs"]["_converted_self_attention"][0],),
                "keep_outputs": True,
            },
        },
    },
    "output_selector": lambda kept: (
        kept["outputs"]["_converted_dropout"], # attention output
        kept["outputs"]["_converted_self_attention"][1], # attention weights
    ),
}
TRANSFORMERS_VILT_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.vilt.modeling_vilt.ViltAttention,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMERS_VILT_ATTENTION,
    }
}

TRANSFORMERS_VILT_OUTPUT = {
    "out_modules": {
        "_converted_dense": {
            "module": lambda in_module: in_module.dense,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (in_args[0],),
                "keep_outputs": True,
            },
        },
        "_converted_dropout": {
            "module": lambda in_module: in_module.dropout,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (kept["outputs"]["_converted_dense"],),
                "keep_outputs": True,
            },
        },
        "_converted_sum": {
            "module": lambda in_module: StackAndSum(),
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_dropout"],
                    in_args[1],
                ),
                "keep_outputs": True,
            },
        },
    },
    "output_selector": lambda kept: kept["outputs"]["_converted_sum"],
}
TRANSFORMERS_VILT_OUTPUT_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.vilt.modeling_vilt.ViltOutput,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMERS_VILT_OUTPUT,
    }
}

TRANSFORMERS_VILT_LAYER = {
    "out_modules": {
        "_converted_layernorm_before": {
            "module": lambda in_module: in_module.layernorm_before,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (in_args[0],),
                "keep_outputs": True,
            },
        },
        "_converted_self_attention": {
            "out_module_type": nn.MultiheadAttention,
            "args": {
                "embed_dim":  lambda in_module: in_module.attention.attention.attention_head_size * in_module.attention.attention.num_attention_heads,
                "num_heads": lambda in_module: in_module.attention.attention.num_attention_heads,
                "dropout": lambda in_module: in_module.attention.attention.dropout.p,
                "bias": True,
                "add_bias_kv": False,
                "add_zero_attn": False,
                "kdim": None,
                "vdim": None,
                "batch_first": True,
                "device": lambda in_module: next(in_module.attention.attention.parameters()).device,
                "dtype": None,
            },
            "params": {
                "in_proj_weight": lambda in_module_params: torch.cat([
                    in_module_params["attention.attention.query.weight"],
                    in_module_params["attention.attention.key.weight"],
                    in_module_params["attention.attention.value.weight"],
                ]),
                "in_proj_bias": lambda in_module_params: torch.cat([
                    in_module_params["attention.attention.query.bias"],
                    in_module_params["attention.attention.key.bias"],
                    in_module_params["attention.attention.value.bias"]
                ]),
                "out_proj.weight": "attention.output.dense.weight",
                "out_proj.bias": "attention.output.dense.bias",
            },
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: tuple(
                    kept["outputs"]["_converted_layernorm_before"] for _ in range(3)
                ),
                "kwargs": {
                    "key_padding_mask": None,
                    "need_weights": False,
                    "attn_mask": transform_attn_mask_of_transformers,
                    "average_attn_weights": True,
                    "is_causal": False,
                },
                "keep_outputs": True,
            },
        },
        "_converted_attention_output": {
            "module": lambda in_module: in_module.attention.output,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_self_attention"][0],
                    kept["outputs"]["_converted_layernorm_before"],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_first_residual_connection": {
            "module": lambda in_module: StackAndSum(),
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_attention_output"],
                    in_args[0],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_layernorm_after": {
            "module": lambda in_module: in_module.layernorm_after,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_first_residual_connection"],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_intermediate": {
            "module": lambda in_module: in_module.intermediate,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_layernorm_after"],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_output_dense": {
            "module": lambda in_module: in_module.output.dense,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_intermediate"],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_output_dropout": {
            "module": lambda in_module: in_module.output.dropout,
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_output_dense"],
                ),
                "keep_outputs": True,
            },
        },
        "_converted_second_residual_connection": {
            "module": lambda in_module: StackAndSum(),
            "forward": {
                "args": lambda in_args, in_kwargs, kept, out_module: (
                    kept["outputs"]["_converted_output_dropout"],
                    kept["outputs"]["_converted_first_residual_connection"],
                ),
                "keep_outputs": True,
            },
        },
    },
    "output_selector": lambda kept: (
        kept["outputs"]["_converted_second_residual_connection"],
        kept["outputs"]["_converted_self_attention"][1],
    ),
}

TRANSFORMERS_VILT_LAYER_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.vilt.modeling_vilt.ViltLayer,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMERS_VILT_LAYER,
    }
}

# transformers lxmert
TRANSFORMERS_LXMERT_SELF_ATTENTION = TRANSFORMERS_BERT_ATTENTION
TRANSFORMERS_LXMERT_SELF_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.lxmert.modeling_lxmert.LxmertSelfAttentionLayer,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMERS_LXMERT_SELF_ATTENTION,
    }
}

TRANSFORMERS_LXMERT_CROSS_ATTENTION = {
    "out_modules": {
        "_converted_cross_attention": {
            "out_module_type": nn.MultiheadAttention,
            "args": {
                "embed_dim":  lambda in_module: in_module.att.attention_head_size * in_module.att.num_attention_heads,
                "num_heads": lambda in_module: in_module.att.num_attention_heads,
                "dropout": lambda in_module: in_module.att.dropout.p,
                "bias": True,
                "add_bias_kv": False,
                "add_zero_attn": False,
                "kdim": None,
                "vdim": None,
                "batch_first": True,
                "device": lambda in_module: next(in_module.att.parameters()).device,
                "dtype": None,
            },
            "params": {
                "q_proj_weight": "att.query.weight",
                "k_proj_weight": "att.key.weight",
                "v_proj_weight": "att.value.weight",
                "in_proj_bias": lambda in_module_params: torch.cat([
                    in_module_params["att.query.bias"],
                    in_module_params["att.key.bias"],
                    in_module_params["att.value.bias"]
                ]),
                "out_proj.weight": "output.dense.weight",
                "out_proj.bias": "output.dense.bias",
            },
            "forward": {
                "args": lambda in_args, in_kwargs, kept: (in_args[0], in_args[1], in_args[1]),
                "kwargs": {
                    "key_padding_mask": None,
                    "need_weights": False,
                    "attn_mask": transform_attn_mask_of_transformers,
                    "average_attn_weights": True,
                    "is_causal": False,
                },
                "keep_outputs": True,
            },
        },
        "_converted_dropout": {
            "module": lambda in_module: in_module.output.dropout,
            "forward": {
                "args": lambda in_args, in_kwargs, kept: (kept["outputs"]["_converted_cross_attention"],),
                "keep_outputs": True,
            },
        },
        "_converted_layernorm": {
            "module": lambda in_module: in_module.output.LayerNorm,
            "forward": {
                "args": lambda in_args, in_kwargs, kept: (
                    kept["outputs"]["_converted_dropout"] + in_args[0],
                ),
                "keep_outputs": True,
            },
        },
    },
    "output_selector": lambda kept: (
        kept["outputs"]["_converted_layernorm"], # attention output
        kept["outputs"]["_converted_cross_attention"][1], # attention weights
    ),
}
TRANSFORMERS_LXMERT_CROSS_ATTENTION_CONVERTER_FACTORY_CONFIG = {
    "in_module_type": transformers.models.lxmert.modeling_lxmert.LxmertCrossAttentionLayer,
    "config_selector": lambda in_module: True,
    "out_module_configs": {
        True: TRANSFORMERS_LXMERT_CROSS_ATTENTION,
    }
}



from typing import Optional, Any
from torch import Tensor
from optuna.trial import Trial, TrialState

from pnpxai.core._types import TensorOrTupleOfTensors
from pnpxai.core.modality.modality import Modality, TextModality
from pnpxai.explainers import Explainer, KernelShap, Lime
from pnpxai.explainers.utils.postprocess import PostProcessor, Identity
from pnpxai.evaluator.metrics.base import Metric
from pnpxai.evaluator.optimizer.suggestor import suggest
from pnpxai.utils import format_into_tuple, format_out_tuple_if_single, generate_param_key


class Objective:
    """
    A class that encapsulates the logic for evaluating a model's performance using a 
    specified explainer, postprocessor, and metric within a given modality. The `Objective` 
    class is designed to be callable and can be used within an optimization framework 
    like Optuna to evaluate different configurations of explainers and postprocessors.

    Parameters:
        explainer (Explainer): 
            The explainer used to generate attributions for the model.
        postprocessor (PostProcessor): 
            The postprocessor applied to the attributions generated by the explainer.
        metric (Metric): 
            The metric used to evaluate the effectiveness of the postprocessed attributions.
        modality (Modality): 
            The data modality (e.g., image, text) the model and explainer are operating on.
        inputs (Optional[TensorOrTupleOfTensors], optional): 
            The input data to the model. Defaults to None.
        targets (Optional[Tensor], optional): 
            The target labels for the input data. Defaults to None.
    """

    EXPLAINER_KEY = 'explainer'
    POSTPROCESSOR_KEY = 'postprocessor'

    def __init__(
        self,
        explainer: Explainer,
        postprocessor: PostProcessor,
        metric: Metric,
        modality: Modality,
        inputs: Optional[TensorOrTupleOfTensors] = None,
        targets: Optional[Tensor] = None,
    ):
        self.explainer = explainer
        self.postprocessor = postprocessor
        self.metric = metric
        self.modality = modality
        self.inputs = inputs
        self.targets = targets

    def set_inputs(self, inputs):
        """
        Sets the input data for the objective.

        Parameters:
            inputs (TensorOrTupleOfTensors): 
                The input data to be used by the explainer.

        Returns:
            Objective: The updated Objective instance.
        """
        self.inputs = inputs
        return self

    def set_targets(self, targets):
        """
        Sets the target labels for the objective.

        Parameters:
            targets (Tensor): 
                The target labels corresponding to the input data.

        Returns:
            Objective: The updated Objective instance.
        """
        self.targets = targets
        return self

    def set_data(self, inputs, targets):
        """
        Sets both the input data and target labels for the objective.

        Parameters:
            inputs (TensorOrTupleOfTensors): 
                The input data to be used by the explainer.
            targets (Tensor): 
                The target labels corresponding to the input data.

        Returns:
            Objective: The updated Objective instance.
        """
        self.set_inputs(inputs)
        self.set_targets(targets)
        return self

    def __call__(self, trial: Trial) -> float:
        """
        Executes the objective function within an optimization trial. This method is 
        responsible for selecting the explainer and postprocessor based on the trial 
        suggestions, performing the explanation and postprocessing, and evaluating the 
        results using the specified metric.

        Parameters:
            trial (Trial): 
                The trial object from an optimization framework like Optuna.

        Returns:
            float: The evaluation result of the metric after applying the explainer and 
            postprocessor to the model's predictions. Returns `nan` if the postprocessed 
            results contain non-countable values like `nan` or `inf`.
        """
        # suggest explainer
        explainer = suggest(trial, self.explainer, self.modality, key=self.EXPLAINER_KEY)

        # suggest postprocessor
        modalities = format_into_tuple(self.modality)
        is_multi_modal = len(modalities) > 1
        postprocessor = []
        for pp, modality in zip(
            format_into_tuple(self.postprocessor),
            modalities,
        ):
            force_params = {}
            if (
                isinstance(explainer, (Lime, KernelShap))
                and isinstance(modality, TextModality)
            ):
                force_params['pooling_fn'] = Identity()
            postprocessor.append(suggest(
                trial, pp, modality,
                key=generate_param_key(
                    self.POSTPROCESSOR_KEY,
                    modality.__class__.__name__ if is_multi_modal else None
                ),
                force_params=force_params,
            ))
        postprocessor = format_out_tuple_if_single(tuple(postprocessor))

        # Ignore duplicated samples
        states_to_consider = (TrialState.COMPLETE,)
        trials_to_consider = trial.study.get_trials(deepcopy=False, states=states_to_consider)
        for t in reversed(trials_to_consider):
            if trial.params == t.params:
                trial.set_user_attr('explainer', explainer)
                trial.set_user_attr('postprocessor', postprocessor)
                return t.value

        # Explain and postprocess
        attrs = explainer.attribute(self.inputs, self.targets)
        postprocessed = tuple(
            pp(attr) for pp, attr in zip(
                format_into_tuple(postprocessor),
                format_into_tuple(attrs),
            )
        )

        if any(pp.isnan().sum() > 0 or pp.isinf().sum() > 0 for pp in postprocessed):
            # Treat a failure as nan
            return float('nan')

        postprocessed = format_out_tuple_if_single(postprocessed)
        metric = self.metric.set_explainer(explainer)
        evals = format_into_tuple(
            metric.evaluate(self.inputs, self.targets, postprocessed)
        )

        # Keep current explainer and postprocessor on trial
        trial.set_user_attr('explainer', explainer)
        trial.set_user_attr('postprocessor', postprocessor)
        return (sum(*evals) / len(evals)).item()



from typing import Tuple
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.grad_cam import GradCam
from pnpxai.explainers.guided_grad_cam import GuidedGradCam
from pnpxai.explainers.gradient import Gradient
from pnpxai.explainers.grad_x_input import GradientXInput
from pnpxai.explainers.smooth_grad import SmoothGrad
from pnpxai.explainers.var_grad import VarGrad
from pnpxai.explainers.integrated_gradients import IntegratedGradients
from pnpxai.explainers.lrp import (
    LRPBase,
    LRPUniformEpsilon,
    LRPEpsilonGammaBox,
    LRPEpsilonPlus,
    LRPEpsilonAlpha2Beta1,
)
from pnpxai.explainers.rap import RAP
from pnpxai.explainers.kernel_shap import KernelShap
from pnpxai.explainers.deep_lift_shap import DeepLiftShap
from pnpxai.explainers.lime import Lime
from pnpxai.explainers.attention_rollout import (
    AttentionRollout,
    TransformerAttribution,
)
from pnpxai.explainers.lear import LEAR
from pnpxai.explainers.gfgp import Gfgp

CAM_BASED_EXPLAINERS = [GradCam, GuidedGradCam]
GRADIENT_BASED_EXPLAINERS = [
    Gradient,
    GradientXInput,
    SmoothGrad,
    VarGrad,
    IntegratedGradients,
    LRPUniformEpsilon,
    LRPEpsilonPlus,
    LRPEpsilonGammaBox,
    LRPEpsilonAlpha2Beta1
]
PERTURBATION_BASED_EXPLAINERS = [
    KernelShap,
    Lime,
    Gfgp
]
ATTENTION_SPECIFIC_EXPLAINERS = [
    AttentionRollout,
    TransformerAttribution,
]
AVAILABLE_EXPLAINERS: Tuple[Explainer] = (
    GradCam,
    GuidedGradCam,
    Gradient,
    GradientXInput,
    SmoothGrad,
    VarGrad,
    IntegratedGradients,
    LRPUniformEpsilon,
    LRPEpsilonPlus,
    LRPEpsilonGammaBox,
    LRPEpsilonAlpha2Beta1,
    RAP,
    KernelShap,
    Lime,
    Gfgp,
    AttentionRollout,
    TransformerAttribution,
)

EXPLAINERS_FOR_TABULAR = []
EXPLAINERS_FOR_MEDICAL_IMAGE: Tuple[Explainer] = (LEAR,)


from pnpxai.core.detector import detect_model_architecture
from pnpxai.core.recommender import XaiRecommender
from pnpxai.core.experiment import (
    Experiment,
    AutoExplanation,
    AutoExplanationForImageClassification,
    AutoExplanationForTextClassification,
    AutoExplanationForVisualQuestionAnswering,
    AutoExplanationForTSClassification,
)


from typing import Tuple, Union, Optional, Dict

from torch import Tensor
from torch.nn.modules import Module
from captum.attr import KernelShap as CaptumKernelShap

from pnpxai.core.detector.types import Linear, Convolution, LSTM, RNN, Attention
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.types import ForwardArgumentExtractor
from pnpxai.explainers.utils.baselines import BaselineMethodOrFunction, BaselineFunction
from pnpxai.explainers.utils.feature_masks import FeatureMaskMethodOrFunction, FeatureMaskFunction
from pnpxai.utils import format_into_tuple, format_out_tuple_if_single


class KernelShap(Explainer):
    """
    KernelSHAP explainer.

    Supported Modules: `Linear`, `Convolution`, `LSTM`, `RNN`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        n_samples (int): Number of samples
        baseline_fn (Union[BaselineMethodOrFunction, Tuple[BaselineMethodOrFunction]]): The baseline function, accepting the attribution input, and returning the baseline accordingly.
        feature_mask_fn (Union[FeatureMaskMethodOrFunction, Tuple[FeatureMaskMethodOrFunction]): The feature mask function, accepting the attribution input, and returning the feature mask accordingly.
        mask_token_id (Optional[int]): The token id of the mask, used for modalities, utilizing tokenization
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Scott M. Lundberg, Su-In Lee. A Unified Approach to Interpreting Model Predictions.
    """

    SUPPORTED_MODULES = [Linear, Convolution, LSTM, RNN, Attention]

    def __init__(
        self,
        model: Module,
        n_samples: int = 25,
        baseline_fn: Union[BaselineMethodOrFunction,
                           Tuple[BaselineMethodOrFunction]] = 'zeros',
        feature_mask_fn: Union[FeatureMaskMethodOrFunction,
                               Tuple[FeatureMaskMethodOrFunction]] = 'felzenszwalb',
        forward_arg_extractor: Optional[ForwardArgumentExtractor] = None,
        additional_forward_arg_extractor: Optional[ForwardArgumentExtractor] = None,
        mask_token_id: Optional[int] = None,
    ) -> None:
        super().__init__(
            model,
            forward_arg_extractor,
            additional_forward_arg_extractor
        )
        self.n_samples = n_samples
        self.baseline_fn = baseline_fn
        self.feature_mask_fn = feature_mask_fn
        self.mask_token_id = mask_token_id

    def attribute(
        self,
        inputs: Tensor,
        targets: Optional[Tensor] = None,
    ) -> Union[Tensor, Tuple[Tensor]]:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor]]: The result of the explanation.
        """
        forward_args, additional_forward_args = self._extract_forward_args(
            inputs)
        forward_args = format_into_tuple(forward_args)
        explainer = CaptumKernelShap(self.model)
        attrs = explainer.attribute(
            inputs=forward_args,
            target=targets,
            baselines=self._get_baselines(forward_args),
            feature_mask=self._get_feature_masks(forward_args),
            n_samples=self.n_samples,
            additional_forward_args=additional_forward_args,
        )
        attrs = format_out_tuple_if_single(attrs)
        return attrs

    def get_tunables(self) -> Dict[str, Tuple[type, Dict]]:
        """
        Provides Tunable parameters for the optimizer

        Tunable parameters:
            `n_samples` (int): Value can be selected in the range of `range(10, 50, 10)`

            `baseline_fn` (callable): BaselineFunction selects suitable values in accordance with the modality

            `feature_mask_fn` (callable): FeatureMaskFunction selects suitable values in accordance with the modality
        """
        return {
            'n_samples': (int, {'low': 10, 'high': 50, 'step': 10}),
            'baseline_fn': (BaselineFunction, {}),
            'feature_mask_fn': (FeatureMaskFunction, {})
        }



from typing import Any, Optional, Union
import torch
from pnpxai.utils import to_device


class ExperimentCache:
    __EXPLAINER_KEY = "explainer"
    __METRIC_KEY = "metric"
    __POSTPROCESSOR_KEY = "postprocessor"
    __EVALUATION_KEY = "evaluation"
    __DATA_KEY = "data"

    def __init__(self, cache_device: Optional[Union[torch.device, str]] = None):
        cpu_device = torch.device("cpu")
        if isinstance(cache_device, str):
            cache_device = torch.device(cache_device)
        self._device = cache_device if cache_device is not None else cpu_device

        self._global_cache = {}

    def to_device(self, x):
        return to_device(x, self._device)

    def get_output(self, data_id: int) -> Optional[Any]:
        key = self._get_key(data_id)
        return self._global_cache.get(key, None)

    def get_explanation(self, data_id: int, explainer_id: int) -> Optional[Any]:
        key = self._get_key(data_id, explainer_id)
        return self._global_cache.get(key, None)

    def get_evaluation(
        self,
        data_id: int,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int
    ) -> Optional[Any]:
        key = self._get_key(data_id, explainer_id, postprocessor_id, metric_id)
        return self._global_cache.get(key, None)

    def set_output(self, data_id: int, output: Any):
        key = self._get_key(data_id)
        self._global_cache[key] = self.to_device(output)

    def set_explanation(self, data_id: int, explainer_id: int, explanation: Any):
        key = self._get_key(data_id, explainer_id)
        self._global_cache[key] = self.to_device(explanation)

    def set_evaluation(
        self,
        data_id: int,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
        evaluation: Any
    ):
        key = self._get_key(data_id, explainer_id, postprocessor_id, metric_id)
        self._global_cache[key] = self.to_device(evaluation)

    def _get_key(
        self,
        data_id: int,
        explainer_id: Optional[int] = None,
        postprocessor_id: Optional[int] = None,
        metric_id: Optional[int] = None
    ):
        key = f"{self.__DATA_KEY}_{data_id}"
        if explainer_id is None:
            return key

        key = f"{key}.{self.__EXPLAINER_KEY}_{explainer_id}"
        if metric_id is None:
            return key

        return f"{key}.{self.__POSTPROCESSOR_KEY}_{postprocessor_id}.{self.__EVALUATION_KEY}_{metric_id}"



from typing import Any, List, Optional, Sequence, Union, Type, Tuple, Callable

import torch
from torch import Tensor
from torch.utils.data import DataLoader, Subset, Dataset

from pnpxai.core.experiment.cache import ExperimentCache
from pnpxai.core._types import DataSource
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils.postprocess import PostProcessor
from pnpxai.evaluator.metrics.base import Metric
from pnpxai.utils import format_into_tuple, format_out_tuple_if_single


class ExperimentManager:
    def __init__(
        self,
        data: DataSource,
        cache_device: Optional[Union[torch.device, str]] = None,
    ):
        self._data = data
        self.set_data_ids()
        self._explainers: List[Explainer] = []
        self._explainer_ids: List[int] = []
        self._postprocessors: List[PostProcessor] =[]
        self._postprocessor_ids: List[int] = []
        self._metrics: List[Metric] = []
        self._metric_ids: List[int] = []

        self._cache = ExperimentCache(cache_device)

    def clear(self):
        self._explainers = []
        self._explainer_ids = []
        self._metrics = []
        self._metric_ids = []
        self._cache._global_cache = {}

    @property
    def data(self):
        return self._data

    @property
    def explainers(self):
        return self._explainers

    @property
    def postprocessors(self):
        return self._postprocessors

    @property
    def metrics(self):
        return self._metrics

    def set_data_ids(self, data_ids: Optional[Sequence[int]] = None):
        self._data_ids = data_ids if data_ids is not None else list(
            range(self._all_data_len))

    def cache_outputs(self, data_ids, outputs):
        assert len(data_ids) == len(outputs)
        for idx, output in zip(data_ids, outputs):
            self._cache.set_output(idx, output)

    def cache_explanations(
        self,
        explainer_id: int,
        data_ids: List[int],
        explanations,
    ):
        explanations = format_into_tuple(explanations)
        assert all(len(expl) == len(data_ids) for expl in explanations)
        for idx, *explanation in zip(data_ids, *explanations):
            explanation = format_into_tuple(explanation)
            self._cache.set_explanation(idx, explainer_id, explanation)

    def cache_evaluations(
        self,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
        data_ids: List[int],
        evaluations: DataSource,
    ):
        evaluations = format_into_tuple(evaluations)
        assert all(len(ev) == len(data_ids) for ev in evaluations)
        for idx, *evaluation in zip(data_ids, *evaluations):
            evaluation = format_into_tuple(evaluation)
            self._cache.set_evaluation(
                idx, explainer_id, postprocessor_id, metric_id, evaluation)

    def get_data_ids(self) -> Sequence[int]:
        return self._data_ids

    def _get_data_by_ids(
        self,
        data_ids: Optional[Sequence[int]]=None,
    ) -> List[Any]:
        if data_ids is None:
            return self._data
        if isinstance(self._data, DataLoader):
            data = self._copy_data_loader(
                data=Subset(self._data.dataset, data_ids),
                data_loader=self._data.__class__
            )
        elif isinstance(self._data, Dataset):
            data = Subset(self._data, data_ids)
        else:
            try:
                data = self._data[data_ids]
            except:
                data = [self._data[idx] for idx in data_ids]
        return data

    def get_data(
        self,
        data_ids: Optional[Sequence[int]]=None,
    ) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids if data_ids is not None else self._data_ids
        return self._get_data_by_ids(data_ids), data_ids

    ###
    def get_data_by_id(self, data_id: int):
        data = self._data.dataset[data_id]
        data = self._data.collate_fn([data])
        return tuple(d.squeeze(0) for d in data)

    def get_output_by_id(self, data_id: int):
        return self._cache.get_output(data_id)

    def get_explainer_by_id(self, explainer_id: int):
        return self._explainers[explainer_id]

    def get_explanation_by_id(self, data_id: int, explainer_id: int):
        cached = self._cache.get_explanation(data_id, explainer_id)
        if isinstance(cached, Tuple):
            return format_out_tuple_if_single(cached)
        return cached # This may be None

    def get_postprocessor_by_id(self, postprocessor_id: int):
        return self._postprocessors[postprocessor_id]

    def get_metric_by_id(self, metric_id: int):
        return self._metrics[metric_id]

    def get_evaluation_by_id(
        self,
        data_id: int,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int
    ):
        cached = self._cache.get_evaluation(
            data_id, explainer_id, postprocessor_id, metric_id)
        if isinstance(cached, Tuple):
            return format_out_tuple_if_single(cached)
        return cached # This may be None

    def batch_data_by_ids(
        self,
        data_ids: List[int],
    ):
        batch = [self._data.dataset[idx] for idx in data_ids]
        batch = self._data.collate_fn(batch)
        return batch

    def batch_outputs_by_ids(self, data_ids: List[int]):
        batch = []
        for data_id in data_ids:
            output = self.get_output_by_id(data_id)
            if output is None:
                raise KeyError(f"Output for {data_id} does not exist in cache.")
            batch.append(output)
        return self._format_batch(batch)

    def batch_explainers_by_ids(self, explainer_ids: List[int]):
        return [self._explainers[idx] for idx in explainer_ids]

    def batch_explanations_by_ids(
        self,
        data_ids: List[int],
        explainer_id: int,
    ):
        batch = [
            self.get_explanation_by_id(data_id, explainer_id)
            for data_id in data_ids
        ]
        return self._format_batch(batch)

    def batch_evaluations_by_ids(
        self,
        data_ids: List[int],
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
    ):
        batch = [
            self.get_evaluation_by_id(
                data_id, explainer_id, postprocessor_id, metric_id)
            for data_id in data_ids
        ]
        return self._format_batch(batch)

    def _format_batch(self, batch):
        if isinstance(batch[0], Tuple):
            cnt = len(batch[0])
            batch = [[data[i] for data in batch] for i in range(cnt)]
        else:
            batch = [batch]
        formatted = *(self._stack_batch(b) for b in batch),
        formatted = format_out_tuple_if_single(formatted)
        return formatted

    def _stack_batch(self, single_batch):
        if torch.is_tensor(single_batch[0]):
            return torch.stack(single_batch)
        return single_batch            

    def add_explainer(self, explainer: Explainer) -> int:
        explainer_id = len(self._explainers)
        self._explainers.append(explainer)
        self._explainer_ids.append(explainer_id)
        return explainer_id

    def _get_explainers_by_ids(self, explainer_ids: Optional[Sequence[int]] = None) -> List[Explainer]:
        # return [self._explainers_w_args[idx] for idx in explainer_ids] if explainer_ids is not None else self._explainers_w_args
        return [self._explainers[idx] for idx in explainer_ids] if explainer_ids is not None else self._explainers

    def get_explainers(self, explainer_ids: Optional[List[int]]=None) -> Tuple[List[Explainer], List[int]]:
        explainer_ids = explainer_ids or self._explainer_ids
        return self._get_explainers_by_ids(explainer_ids), explainer_ids

    def add_postprocessor(self, postprocessor: Callable) -> int:
        postprocessor_id = len(self._postprocessors)
        self._postprocessors.append(postprocessor)
        self._postprocessor_ids.append(postprocessor_id)
        return postprocessor_id

    def _get_postprocessors_by_ids(self, postprocessor_ids: Optional[Sequence[int]]=None) -> List[Callable]:
        return [self._postprocessors[idx] for idx in postprocessor_ids] if postprocessor_ids is not None else self._postprocessors

    def get_postprocessors(self, postprocessor_ids: Optional[List[int]]=None) -> Tuple[List[Callable], List[int]]:
        postprocessor_ids = postprocessor_ids or self._postprocessor_ids
        return self._get_postprocessors_by_ids(postprocessor_ids), postprocessor_ids

    # metric
    def add_metric(self, metric: Metric) -> int:
        metric_id = len(self._metrics)
        self._metrics.append(metric)
        self._metric_ids.append(metric_id)
        return metric_id

    def _get_metrics_by_ids(self, metric_ids: Optional[Sequence[int]] = None) -> List[Metric]:
        return [self._metrics[idx] for idx in metric_ids] if metric_ids is not None else self._metrics

    def get_metrics(self, metric_ids: Optional[List[int]]=None) -> Tuple[List[Metric], List[int]]:
        metric_ids = metric_ids or self._metric_ids
        return self._get_metrics_by_ids(metric_ids), metric_ids

    # explanation
    def get_data_to_process_for_explainer(
        self,
        explainer_id: int,
        data_ids: Optional[List[int]] = None,
    ) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids or self._data_ids
        data_ids = [
            idx for idx in data_ids if self._cache.get_explanation(idx, explainer_id) is None
        ]
        return self._get_data_by_ids(data_ids), data_ids

    def save_explanations(self, explanations: DataSource, data: DataSource, data_ids: Sequence[int], explainer_id: int):
        explanations = self.flatten_if_batched(explanations, data)
        assert len(explanations) == len(data_ids)
        for idx, explanation in zip(data_ids, explanations):
            self._cache.set_explanation(idx, explainer_id, explanation)

    def get_data_to_process_for_metric(
        self,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
        data_ids: Optional[List[int]]=None,
    ) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids or self._data_ids
        data_ids = [
            idx for idx in data_ids
            if self._cache.get_evaluation(idx, explainer_id, postprocessor_id, metric_id) is None
        ]
        return self._get_data_by_ids(data_ids), data_ids

    def get_data_to_predict(self, data_ids: List[int]) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids or self._data_ids
        data_ids = [
            idx for idx in data_ids if self._cache.get_output(idx) is None
        ]
        return self._get_data_by_ids(data_ids), data_ids


    def get_all_data(self) -> DataSource:
        return self._get_data_by_ids()

    def get_valid_explanations(self, explainer_id: int, data_ids: Optional[Sequence[int]] = None) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids if data_ids is not None else self._data_ids
        explanations = self.get_flat_explanations(explainer_id, data_ids)
        valid_explanations = []
        valid_data_ids = []

        for idx, explanation in zip(data_ids, explanations):
            if explanation is None:
                continue
            valid_explanations.append(explanation)
            valid_data_ids.append(idx)

        if self.is_batched:
            return self._copy_data_loader(valid_explanations, self._data.__class__, do_collate=False), valid_data_ids

        return valid_explanations, valid_data_ids

    def get_flat_explanations(self, explainer_id: int, data_ids: Optional[Sequence[int]] = None) -> Sequence[Tensor]:
        data_ids = data_ids if data_ids is not None else self._data_ids
        return [self._cache.get_explanation(idx, explainer_id) for idx in data_ids]

    def get_flat_evaluations(
        self,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
        data_ids: Optional[Sequence[int]] = None
    ) -> Sequence[Tensor]:
        data_ids = data_ids if data_ids is not None else self._data_ids
        return [self._cache.get_evaluation(
            idx, explainer_id, postprocessor_id, metric_id) for idx in data_ids]

    def get_flat_outputs(self, data_ids: Optional[Sequence[int]] = None) -> Tuple[DataSource, List[int]]:
        data_ids = data_ids if data_ids is not None else self._data_ids
        return [self._cache.get_output(idx) for idx in data_ids]

    def save_evaluations(
        self,
        evaluations: DataSource,
        data: DataSource,
        data_ids: Sequence[int],
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int
    ):
        evaluations = self.flatten_if_batched(evaluations, data)
        assert len(evaluations) == len(data_ids)

        for idx, evaluation in zip(data_ids, evaluations):
            self._cache.set_evaluation(
                idx, explainer_id, metric_id, postprocessor_id, evaluation)

    def save_outputs(self, outputs: DataSource, data: DataSource, data_ids: Sequence[int]):
        outputs = self.flatten_if_batched(outputs, data)
        assert len(outputs) == len(data_ids)

        for idx, output in zip(data_ids, outputs):
            self._cache.set_output(idx, output)


    def _get_batch_size(self, data: DataSource) -> Optional[int]:
        if torch.is_tensor(data):
            return len(data)
        for item in data:
            item_len = self._get_batch_size(item)
            if item_len is not None:
                return item_len
        return None

    def flatten_if_batched(self, values: DataSource, length_reference: DataSource):
        if not self.is_batched:
            return values

        assert len(values) == len(length_reference)

        flattened = []
        for batch, reference in zip(values, length_reference):
            if batch is None:
                reference_len = self._get_batch_size(reference) or 1
                batch = [None] * reference_len
            if isinstance(batch, Tuple):
                batch = zip(*[list(b) for b in batch])
            flattened.extend(list(batch))
        return flattened

    def _copy_data_loader(self, data: DataSource, data_loader: Type[DataLoader], do_collate=True):
        duplicated_params = [
            'batch_size', 'num_workers', 'pin_memory', 'drop_last', 'timeout',
            'worker_init_fn', 'multiprocessing_context', 'generator', 'persistent_workers', 'pin_memory_device'
        ]
        collate_fn = getattr(self._data, 'collate_fn') if do_collate else None
        return data_loader(
            dataset=data, shuffle=False, collate_fn=collate_fn,
            **{param: getattr(self._data, param) for param in duplicated_params}
        )

    @property
    def all_explainers(self) -> Sequence[Explainer]:
        # return self._explainers_w_args
        return self._explainers

    @property
    def all_metrics(self) -> Sequence[Metric]:
        return self._metrics

    @property
    def has_metrics(self):
        return len(self._metrics) > 0

    @property
    def has_explanations(self):
        for explainer_id in self._explainer_ids:
            _, valid_data_ids = self.get_valid_explanations(explainer_id)
            if len(valid_data_ids) > 0:
                return True
        return False

    @property
    def is_batched(self):
        return isinstance(self._data, DataLoader) and self._data.batch_size is not None

    @property
    def _all_data_len(self):
        return len(self._data.dataset if isinstance(self._data, DataLoader) else self._data)

    @property
    def _all_data_ids(self):
        return range(self._all_data_len)

    @property
    def _available_data_ids(self):
        return self._all_data_ids if self._data_ids is None else self._data_ids



from typing import Any, Callable, Optional, Sequence, Union, List, Literal

import torch
from torch import Tensor
import numpy as np
import optuna

from pnpxai.core._types import DataSource, Model
from pnpxai.core.modality.modality import Modality, TextModality
from pnpxai.core.experiment.experiment_metrics_defaults import EVALUATION_METRIC_REVERSE_SORT, EVALUATION_METRIC_SORT_PRIORITY
from pnpxai.core.experiment.manager import ExperimentManager
from pnpxai.explainers import Explainer, Lime, KernelShap
from pnpxai.explainers.utils.postprocess import Identity
from pnpxai.evaluator.optimizer.types import OptimizationOutput
from pnpxai.evaluator.optimizer.objectives import Objective
from pnpxai.evaluator.optimizer.utils import (
    load_sampler,
    get_default_n_trials,
)
from pnpxai.evaluator.metrics.base import Metric

from pnpxai.utils import (
    class_to_string, Observable, to_device,
    format_into_tuple, format_out_tuple_if_single,
)


def default_input_extractor(x):
    return x[0]


def default_label_extractor(x):
    return x[-1]


def default_target_extractor(y):
    return y.argmax(-1)


class Experiment(Observable):
    """
    A class representing an experiment for model interpretability.

    Args:
        model (Model): The machine learning model to be interpreted.
        data (DataSource): The data used for the experiment.
        modality (Modality): The type of modality (image, tabular, text, time series) the model is designed for.
        explainers (Sequence[Explainer]): Explainer objects or their arguments for interpreting the model.
        postprocessors (Optional[Sequence[Callable]]): Postprocessing functions to apply to explanations.
        metrics (Optional[Sequence[Metric]]): Evaluation metrics used to assess model interpretability.
        input_extractor (Optional[Callable[[Any], Any]]): Function to extract inputs from data.
        label_extractor (Optional[Callable[[Any], Any]]): Function to extract labels from data.
        target_extractor (Optional[Callable[[Any], Any]]): Function to extract targets from data.
        cache_device (Optional[Union[torch.device, str]]): Device to cache data and results.
        target_labels (bool): True if the target is a label, False otherwise.

    Attributes:
        modality (Modality): Object defining the modality-specific control flow of the experiment.
        manager (ExperimentManager): Manager object for the experiment.
        all_explainers (Sequence[Explainer]): All explainer objects used in the experiment.
        all_metrics (Sequence[Metric]): All evaluation metrics used in the experiment.
        errors (Sequence[Error]): 
        is_image_task (bool): True if the modality is an image-related modality, False otherwise.
        has_explanations (bool): True if the experiment has explanations, False otherwise.
    """

    def __init__(
        self,
        model: Model,
        data: DataSource,
        modality: Modality,
        explainers: Sequence[Explainer],
        postprocessors: Sequence[Callable],
        metrics: Sequence[Metric],
        input_extractor: Optional[Callable[[Any], Any]] = None,
        label_extractor: Optional[Callable[[Any], Any]] = None,
        target_extractor: Optional[Callable[[Any], Any]] = None,
        cache_device: Optional[Union[torch.device, str]] = None,
        target_labels: bool = False,
    ):
        super(Experiment, self).__init__()
        self.model = model
        self.model_device = next(self.model.parameters()).device

        self.manager = ExperimentManager(data=data, cache_device=cache_device)
        for explainer in explainers:
            self.manager.add_explainer(explainer)
        for postprocessor in postprocessors:
            self.manager.add_postprocessor(postprocessor)
        for metric in metrics:
            self.manager.add_metric(metric)

        self.input_extractor = input_extractor \
            if input_extractor is not None \
            else default_input_extractor
        self.label_extractor = label_extractor \
            if label_extractor is not None \
            else default_target_extractor
        self.target_extractor = target_extractor \
            if target_extractor is not None \
            else default_target_extractor
        self.target_labels = target_labels
        self.modality = modality
        self.reset_errors()

    def reset_errors(self):
        self._errors: List[BaseException] = []

    @property
    def errors(self):
        return self._errors

    def to_device(self, x):
        return to_device(x, self.model_device)

    def run_batch(
        self,
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
        data_ids: Optional[Sequence[int]] = None,
    ) -> dict:
        """
        Runs the experiment for selected batch of data, explainer, postprocessor and metric.

        Args:
            data_ids (Sequence[int]): A sequence of data IDs to specify the subset of data to process.
            explainer_id (int): ID of explainer to use for the run.
            postprocessor_id (int): ID of postprocessor to use for the run.
            metrics_id (int): ID of metric to use for the run.

        Returns:
            The dictionary of inputs, labels, outputs, targets, explainer, explanation, postprocessor, postprocessed, metric, and evaluation.

        This method orchestrates the experiment by configuring the manager, obtaining explainer and metric instances,
        processing data, generating explanations, and evaluating metrics. It then caches the results in the manager, and returns back to the user.

        Note: The input parameters allow for flexibility in specifying subset of data, explainer, postprocessor and metric to process.
        """
        data_ids = data_ids if data_ids is not None else self.manager.get_data_ids(
            data_ids)

        self.predict_batch(data_ids)
        self.explain_batch(data_ids, explainer_id)
        self.evaluate_batch(
            data_ids, explainer_id, postprocessor_id, metric_id)
        data = self.manager.batch_data_by_ids(data_ids)
        return {
            'inputs': self.input_extractor(data),
            'labels': self.label_extractor(data),
            'outputs': self.manager.batch_outputs_by_ids(data_ids),
            'targets': self._get_targets(data_ids),
            'explainer': self.manager.get_explainer_by_id(explainer_id),
            'explanation': self.manager.batch_explanations_by_ids(data_ids, explainer_id),
            'postprocessor': self.manager.get_postprocessor_by_id(postprocessor_id),
            'postprocessed': self.postprocess_batch(data_ids, explainer_id, postprocessor_id),
            'metric': self.manager.get_metric_by_id(metric_id),
            'evaluation': self.manager.batch_evaluations_by_ids(data_ids, explainer_id, postprocessor_id, metric_id),
        }

    def predict_batch(
        self,
        data_ids: Sequence[int],
    ):
        """
        Predicts results of the experiment for selected batch of data.

        Args:
            data_ids (Sequence[int]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Batched model outputs corresponding to data ids.

        This method orchestrates the experiment by configuring the manager and processing data. It then caches the results in the manager, and returns back to the user.
        """

        data_ids_pred = [
            idx for idx in data_ids
            if self.manager.get_output_by_id(idx) is None
        ]
        if len(data_ids_pred) > 0:
            data = self.manager.batch_data_by_ids(data_ids_pred)
            outputs = self.model(
                *format_into_tuple(self.input_extractor(data)))
            self.manager.cache_outputs(data_ids_pred, outputs)
        return self.manager.batch_outputs_by_ids(data_ids)

    def explain_batch(
        self,
        data_ids: Sequence[int],
        explainer_id: int,
    ):
        """
        Explains selected batch of data within experiment.

        Args:
            data_ids (Sequence[int]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Batched model explanations corresponding to data ids.

        This method orchestrates the experiment by configuring the manager, obtaining explainer instance,
        processing data, and generating explanations. It then caches the results in the manager, and returns back to the user.
        """
        data_ids_expl = [
            idx for idx in data_ids
            if self.manager.get_explanation_by_id(idx, explainer_id) is None
        ]
        if len(data_ids_expl):
            data = self.manager.batch_data_by_ids(data_ids_expl)
            inputs = self.input_extractor(data)
            targets = self._get_targets(data_ids_expl)
            explainer = self.manager.get_explainer_by_id(explainer_id)
            explanations = explainer.attribute(inputs, targets)
            self.manager.cache_explanations(
                explainer_id, data_ids_expl, explanations)
        return self.manager.batch_explanations_by_ids(data_ids, explainer_id)

    def postprocess_batch(
        self,
        data_ids: List[int],
        explainer_id: int,
        postprocessor_id: int,
    ):
        """
        Postprocesses selected batch of data within experiment.

        Args:
            data_ids (Sequence[int]): A sequence of data IDs to specify the subset of data to postprocess.
            explainer_id (int): An explainer ID to specify the explainer to use.
            postprocessor_id (int): A postprocessor ID to specify the postprocessor to use.

        Returns:
            Batched postprocessed model explanations corresponding to data ids.

        This method orchestrates the experiment by configuring the manager, obtaining explainer instance,
        processing data, and generating explanations. It then caches the results in the manager, and returns back to the user.
        """
        explanations = self.manager.batch_explanations_by_ids(
            data_ids, explainer_id)
        postprocessor = self.manager.get_postprocessor_by_id(postprocessor_id)

        modalities = format_into_tuple(self.modality)
        explanations = format_into_tuple(explanations)
        postprocessors = format_into_tuple(postprocessor)

        batch = []
        explainer = self.manager.get_explainer_by_id(explainer_id)
        for mod, attr, pp in zip(modalities, explanations, postprocessors):
            if (
                isinstance(explainer, (Lime, KernelShap))
                and isinstance(mod, TextModality)
                and not isinstance(pp.pooling_fn, Identity)
            ):
                raise ValueError(
                    f'postprocessor {postprocessor_id} does not support explainer {explainer_id}.')
            batch.append(pp(attr))
        return format_out_tuple_if_single(batch)

    def evaluate_batch(
        self,
        data_ids: List[int],
        explainer_id: int,
        postprocessor_id: int,
        metric_id: int,
    ):
        """
        Evaluates selected batch of data within experiment.

        Args:
            data_ids (Sequence[int]): A sequence of data IDs to specify the subset of data to postprocess.
            explainer_id (int): An explainer ID to specify the explainer to use.
            postprocessor_id (int): A postprocessor ID to specify the postprocessor to use.
            metric_id (int): A metric ID to evaluate the model explanations.

        Returns:
            Batched model evaluations corresponding to data ids.

        This method orchestrates the experiment by configuring the manager, obtaining explainer instance,
        processing data, generating explanations, and evaluating results. It then caches the results in the manager, and returns back to the user.
        """
        data_ids_eval = [
            idx for idx in data_ids
            if self.manager.get_evaluation_by_id(
                idx, explainer_id, postprocessor_id, metric_id) is None
        ]
        if len(data_ids_eval):
            data = self.manager.batch_data_by_ids(data_ids_eval)
            inputs = self.input_extractor(data)
            targets = self._get_targets(data_ids_eval)
            postprocessed = self.postprocess_batch(
                data_ids_eval, explainer_id, postprocessor_id)
            explainer = self.manager.get_explainer_by_id(explainer_id)
            metric = self.manager.get_metric_by_id(metric_id)
            evaluations = metric.set_explainer(explainer).evaluate(
                inputs, targets, postprocessed)
            self.manager.cache_evaluations(
                explainer_id, postprocessor_id, metric_id,
                data_ids_eval, evaluations
            )
        return self.manager.batch_evaluations_by_ids(
            data_ids, explainer_id, postprocessor_id, metric_id
        )

    def _get_targets(self, data_ids):
        if self.target_labels:
            return self.label_extractor(self.manager.batch_data_by_ids(data_ids))
        outputs = self.manager.batch_outputs_by_ids(data_ids)
        return self.target_extractor(outputs)

    def optimize(
        self,
        data_ids: Union[int, Sequence[int]],
        explainer_id: int,
        metric_id: int,
        direction: Literal['minimize', 'maximize'] = 'maximize',
        sampler: Literal['grid', 'random', 'tpe'] = 'tpe',
        n_trials: Optional[int] = None,
        timeout: Optional[float] = None,
        **kwargs,  # sampler kwargs
    ):
        """
        Optimize experiment hyperparameters by processing data, generating explanations, evaluating with metrics, caching and retrieving the data.

        Args:
            data_ids (Union[int, Sequence[int]]): A single data ID or sequence of data IDs to specify the subset of data to process.
            explainer_id (int): An explainer ID to specify the explainer to use.
            metric_id (int): A metric ID to evaluate optimizer decisions.
            direction (Literal['minimize', 'maximize']): A string to specify the direction of optimization.
            sampler (Literal['grid', 'random', 'tpe']): A string to specify the sampler to use for optimization.
            n_trials (Optional[int]): An integer to specify the number of trials for optimization. If none passed, the number of trials is inferred from `timeout`.
            timeout (Optional[float]): A float to specify the timeout for optimization. Ignored, if `n_trials` is specified.

        Returns:
            The Experiment instance with updated results and state.

        This method orchestrates the experiment by configuring the manager, obtaining explainer and metric instances,
        processing data, generating explanations, and evaluating metrics. It then saves the results in the manager.

        Note: The input parameters allow for flexibility in specifying subsets of data, explainers, and metrics to process.
        If not provided, the method processes all available data, explainers, postprocessors, and metrics.
        """
        data_ids = [data_ids] if isinstance(data_ids, int) else data_ids
        data = self.manager.batch_data_by_ids(data_ids)
        explainer = self.manager.get_explainer_by_id(explainer_id)
        postprocessor = self.manager.get_postprocessor_by_id(
            0)  # sample postprocessor to ensure channel_dim
        metric = self.manager.get_metric_by_id(metric_id)

        objective = Objective(
            explainer=explainer,
            postprocessor=postprocessor,
            metric=metric,
            modality=self.modality,
            inputs=self.input_extractor(data),
            targets=self._get_targets(data_ids),
        )
        # TODO: grid search
        if timeout is None:
            n_trials = n_trials or get_default_n_trials(sampler)

        # optimize
        study = optuna.create_study(
            sampler=load_sampler(sampler, **kwargs),
            direction=direction,
        )
        study.optimize(
            objective,
            n_trials=n_trials,
            timeout=timeout,
            n_jobs=1,
        )
        opt_explainer = study.best_trial.user_attrs['explainer']
        opt_postprocessor = study.best_trial.user_attrs['postprocessor']
        return OptimizationOutput(
            explainer=opt_explainer,
            postprocessor=opt_postprocessor,
            study=study,
        )

    def get_inputs_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Tensor]:
        """
        Retrieve and flatten last run input data.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened input data.

        This method retrieves input data using the input extractor and flattens it for further processing.

        Note: The input parameters allow for flexibility in specifying subsets of data to process. If not provided, the method processes all available data.
        """
        data, _ = self.manager.get_data(data_ids)
        data = [self.input_extractor(datum) for datum in data]
        return self.manager.flatten_if_batched(data, data)

    def get_all_inputs_flattened(self) -> Sequence[Tensor]:
        """
        Retrieve and flatten all input data.

        Returns:
            Flattened input data from all available data.

        This method retrieves input data from all available data points using the input extractor and flattens it.
        """
        data = self.manager.get_all_data()
        data = [self.input_extractor(datum) for datum in data]
        return self.manager.flatten_if_batched(data, data)

    def get_labels_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Tensor]:
        """
        Retrieve and flatten labels data.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened labels data.

        This method retrieves label data using the label extractor and flattens it for further processing.
        """
        data, _ = self.manager.get_data(data_ids)
        labels = [self.label_extractor(datum) for datum in data]
        return self.manager.flatten_if_batched(labels, data)

    def get_targets_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Tensor]:
        """
        Retrieve and flatten target data.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened target data.

        This method retrieves target data using the target extractor and flattens it for further processing.

        Note: The input parameters allow for flexibility in specifying subsets of data to process. If not provided, the method processes all available data.
        """
        if self.target_labels:
            return self.get_labels_flattened(data_ids)
        data, _ = self.manager.get_data(data_ids)
        targets = [self._get_targets(data_ids)]
        return self.manager.flatten_if_batched(targets, data)
        # targets = [self.label_extractor(datum) for datum in data] \
        #     if self.target_labels else [self._get_targets(data_ids)]
        # return self.manager.flatten_if_batched(targets, data)

    def get_outputs_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Tensor]:
        """
        Retrieve and flatten model outputs.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened model outputs.

        This method retrieves flattened model outputs using the manager's get_flat_outputs method.

        Note: The input parameters allow for flexibility in specifying subsets of data to process. If not provided, the method processes all available data.
        """
        return self.manager.get_flat_outputs(data_ids)

    def get_explanations_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Sequence[Tensor]]:
        """
        Retrieve and flatten explanations from all explainers.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened explanations from all explainers.

        This method retrieves flattened explanations for each explainer using the manager's `get_flat_explanations` method.

        Note: The input parameters allow for flexibility in specifying subsets of data to process. If not provided, the method processes all available data.
        """
        _, explainer_ids = self.manager.get_explainers()
        return [
            self.manager.get_flat_explanations(explainer_id, data_ids)
            for explainer_id in explainer_ids
        ]

    def get_evaluations_flattened(self, data_ids: Optional[Sequence[int]] = None) -> Sequence[Sequence[Sequence[Tensor]]]:
        """
        Retrieve and flatten evaluations for all explainers and metrics.

        Args:
            data_ids (Optional[Sequence[int]]): A sequence of data IDs to specify the subset of data to process.

        Returns:
            Flattened evaluations for all explainers and metrics.

        This method retrieves flattened evaluations for each explainer and metric using the manager's
        get_flat_evaluations method.

        Note: The input parameters allow for flexibility in specifying subsets of data to process. If not provided, the method processes all available data.
        """
        _, explainer_ids = self.manager.get_explainers()
        _, postprocessor_ids = self.manager.get_postprocessors()
        _, metric_ids = self.manager.get_metrics()

        formatted = [[[
            self.manager.get_flat_evaluations(
                explainer_id, postprocessor_id, metric_id, data_ids)
            for metric_id in metric_ids
        ] for postprocessor_id in postprocessor_ids]
            for explainer_id in explainer_ids]

        return formatted

    def get_explainers_ranks(self) -> Optional[Sequence[Sequence[int]]]:
        """
        Calculate and return rankings for explainers based on evaluations.

        Returns:
            Rankings of explainers. Returns None if rankings cannot be calculated.

        This method calculates rankings for explainers based on evaluations and metric scores. It considers
        metric priorities and sorting preferences to produce rankings.
        """
        evaluations = [[[
            data_evaluation.detach().cpu() if data_evaluation is not None else None
            for data_evaluation in metric_data
        ]for metric_data in explainer_data
        ]for explainer_data in self.get_evaluations_flattened()]
        # (explainers, metrics, data)
        evaluations = np.array(evaluations, dtype=float)
        evaluations = np.nan_to_num(evaluations, nan=-np.inf)
        if evaluations.ndim != 3:
            return None

        evaluations = evaluations.argsort(axis=-3).argsort(axis=-3) + 1
        n_explainers = evaluations.shape[0]
        metric_name_to_idx = {}

        for idx, metric in enumerate(self.get_current_metrics()):
            metric_name = class_to_string(metric)
            evaluations[:, idx, :] = evaluations[:, idx, :]
            if EVALUATION_METRIC_REVERSE_SORT.get(metric_name, False):
                evaluations[:, idx, :] = \
                    n_explainers - evaluations[:, idx, :] + 1

            metric_name_to_idx[metric_name] = idx
        # (explainers, data)
        scores: np.ndarray = evaluations.sum(axis=-2)

        for metric_name in EVALUATION_METRIC_SORT_PRIORITY:
            if metric_name not in metric_name_to_idx:
                continue

            idx = metric_name_to_idx[metric_name]
            scores = scores * n_explainers + evaluations[:, idx, :]

        return scores.argsort(axis=-2).argsort(axis=-2).tolist()

    @property
    def has_explanations(self):
        return self.manager.has_explanations



__all__ = [
    "_assert_position_embedding_type_is_absolute_for_bert_self_attention_of_transformers",
    "_assert_head_mask_is_none_for_bert_attention_of_transformers",
    "_assert_is_not_decoder_for_bert_self_attention_of_transformers",
]

def _assert_position_embedding_type_is_absolute_for_bert_self_attention_of_transformers(module):
    assert module.position_embedding_type == "absolute", f"module conversion for position_embedding_type='{module.position_embedding_type}' not supported."

def _assert_head_mask_is_none_for_bert_attention_of_transformers(args, kwargs):
    head_mask = kwargs["head_mask"] or args[2]
    assert head_mask is None

def _assert_is_not_decoder_for_bert_self_attention_of_transformers(module):
    assert not module.is_decoder



import torch
from torch.nn.modules import Module
from zennit.layer import Sum

class StackAndSum(Module):
    def __init__(self, dim=-1) -> None:
        super().__init__()
        self.dim = dim
        self.sum = Sum()
    
    def forward(self, x, y):
        out = torch.stack([x, y], dim=self.dim)
        out = self.sum(out)
        return out


from typing import Optional

import torch
import numpy as np
from torchvision.transforms.functional import rgb_to_grayscale
from scipy.stats import entropy

from pnpxai.core._types import Model
from pnpxai.explainers.base import Explainer
from pnpxai.evaluator.metrics.base import Metric


class Complexity(Metric):
    """
    Computes the complexity of attributions.

    Given `attributions`, calculates a fractional contribution distribution `prob_mass`,
    ``prob_mass[i] = hist[i] / sum(hist)``. where ``hist[i] = histogram(attributions[i])``.

    The complexity is defined by the entropy,
    ``evaluation = -sum(hist * ln(hist))``


    Args:
        model (Model): The model used for evaluation
        explainer (Optional[Explainer]): The explainer used for evaluation.
        n_bins (int): The number of bins for histogram computation.

    Reference:
        U. Bhatt, A. Weller, and J. M. F. Moura. Evaluating and aggregating feature-based model attributions. In Proceedings of the IJCAI (2020).
    """

    def __init__(
        self, model: Model, explainer: Optional[Explainer] = None, n_bins: int = 10
    ):
        super().__init__(model, explainer)
        self.n_bins = n_bins

    def evaluate(
        self,
        inputs: torch.Tensor,
        targets: torch.Tensor,
        attributions: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Evaluate the explainer's complexity based on their probability masses.

        Args:
            inputs (Optional[Tensor]): The input tensors to the model.
            targets (Optional[Tensor]): The target labels for the inputs.
            attributions (Optional[Tensor]): The attributions for the inputs.

        Returns:
            Tensor: A tensor of the complexity evaluations.
        """
        attributions = self._get_attributions(inputs, targets, attributions)

        assert attributions.ndim in [3, 4], "Must have 2D or 3D attributions"
        if attributions.ndim == 4:
            attributions = rgb_to_grayscale(attributions)
        evaluations = []
        for attr in attributions:
            hist, _ = np.histogram(attr.detach().cpu(), bins=self.n_bins)
            prob_mass = hist / hist.sum()
            evaluations.append(entropy(prob_mass))
        return torch.tensor(evaluations)



from typing import Callable, Optional, Tuple, Union

from torch import Tensor
from torch.nn import Module

from pnpxai.explainers.base import Explainer
from pnpxai.utils import format_into_tuple
from pnpxai.explainers.base import Explainer


class ZennitExplainer(Explainer):
    def __init__(
        self,
        model: Module,
        forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
        additional_forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
        n_classes: Optional[int]=None,
        **kwargs
    ) -> None:
        super().__init__(model, forward_arg_extractor, additional_forward_arg_extractor)
        self.n_classes = n_classes


    def __init_subclass__(cls) -> None:
        cls.attribute = set_n_classes_before(cls.attribute)
        return super().__init_subclass__()


def set_n_classes_before(func):
    def wrapper(*args, **kwargs):
        self = args[0]
        inputs = kwargs.get('inputs', None)
        if inputs is None:
            inputs = args[1]
        if isinstance(inputs, Tensor):
            inputs = format_into_tuple(inputs)
        if self.n_classes is None:
            outputs = self.model(*inputs)
            self.n_classes = outputs.shape[-1]
        return func(*args, **kwargs)
    return wrapper


import torch
from transformers import AutoModel
from pnpxai.explainers.base import Explainer


class LEAR(Explainer):
    """
    LEAR Explainer class for generating counterfactual explanations.

    Args:
        model_name (str): The name of the pre-trained model to load.
        device (str, optional): Device to use for computation ('cuda' or 'cpu').
        model (torch.nn.Module, optional): This is not used for LEAR. It is included for compatibility with the base class.
    """

    def __init__(self, model=None, model_name: str="wltjr1007/LEAR", device: torch.device=None):
        super().__init__(model=model)
        self.device = device or ('cuda' if torch.cuda.is_available() else 'cpu')
        # Load pre-trained model for counterfactual generation (trust_remote_code=True allows custom code from model repository)
        self.CMG = AutoModel.from_pretrained(model_name, trust_remote_code=True).to(self.device)
        self.num_labels = self.CMG.config.num_labels
        self.CMG.eval()

    def attribute(self, inputs, targets):
        """
        Generate counterfactual images by attributing changes to input features.

        Args:
            inputs (torch.Tensor): Input tensor containing original images.
            targets (torch.Tensor): Labels tensor specifying target classes for counterfactuals.

        Returns:
            torch.Tensor: Counterfactual images generated by the model.
        """
        inputs = inputs.to(self.device)

        if targets.dim() == 1:
            targets = torch.nn.functional.one_hot(targets, num_classes=self.num_labels)
        targets = targets.to(self.device)

        with torch.inference_mode():
            counterfactual = self.CMG(inputs, targets)
        return counterfactual


from typing import Tuple, Callable, Sequence, Union, Optional

from torch import Tensor
from torch.nn.modules import Module

from pnpxai.core.detector.types import Linear, Convolution, LSTM, RNN, Attention
from pnpxai.explainers.zennit.attribution import Gradient as GradientAttributor
from pnpxai.explainers.zennit.attribution import LayerGradient as LayerGradientAttributor
from pnpxai.explainers.zennit.base import ZennitExplainer
from pnpxai.explainers.utils import captum_wrap_model_input


class Gradient(ZennitExplainer):
    """
    Gradient explainer.

    Supported Modules: `Linear`, `Convolution`, `LSTM`, `RNN`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
		layer (Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]): The target module to be explained.
		n_classes (int): The number of classes.
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Gabriel Erion, Joseph D. Janizek, Pascal Sturmfels, Scott Lundberg, Su-In Lee. Improving performance of deep learning models with axiomatic attribution priors and expected gradients.
    """
    SUPPORTED_MODULES = [Linear, Convolution, LSTM, RNN, Attention]

    def __init__(
        self,
        model: Module,
        forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]]=None,
        additional_forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]]=None,
        layer: Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]=None,
        n_classes: Optional[int] = None,
    ) -> None:
        super().__init__(
            model,
            forward_arg_extractor,
            additional_forward_arg_extractor,
            n_classes
        )
        self.layer = layer

    @property
    def _layer_attributor(self) -> LayerGradientAttributor:
        wrapped_model = captum_wrap_model_input(self.model)
        layers = [
            wrapped_model.input_maps[layer] if isinstance(layer, str)
            else layer for layer in self.layer
        ] if isinstance(self.layer, Sequence) else [self.layer]
        if len(layers) == 1:
            layers = layers[0]
        return LayerGradientAttributor(
            model=wrapped_model,
            layer=layers,
        )

    @property
    def _attributor(self) -> GradientAttributor:
        return GradientAttributor(self.model)
    
    @property
    def attributor(self) -> Union[GradientAttributor, LayerGradientAttributor]:
        if self.layer is None:
            return self._attributor
        return self._layer_attributor

    def attribute(
        self,
        inputs: Union[Tensor, Tuple[Tensor]],
        targets: Tensor
    ) -> Union[Tensor, Tuple[Tensor]]:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor]]: The result of the explanation.
        """
        forward_args, additional_forward_args = self._extract_forward_args(inputs)
        attrs = self.attributor.forward(
            forward_args,
            targets,
            additional_forward_args,
        )
        return attrs


from torch.utils.data import Dataset, DataLoader
from torch import Tensor, nn
from typing import Union, Sequence, Literal, Tuple
from enum import Enum


TensorOrTupleOfTensors = Union[Tensor, Tuple[Tensor]]
TensorSequence = Union[Dataset, Sequence[Tensor]]
TensorOrTensorSequence = Union[TensorSequence, Tensor]
DataSource = Union[DataLoader, TensorOrTensorSequence]

Model = nn.Module  # TODO: List other model types in Union[Type1, Type2, ...]
Task = Literal["classification"]
ExplanationType = Literal["attribution"]

class ConfigKeys(Enum):
    EXPLAINERS = 'explainers'
    METRICS = 'metrics'
    TASK = 'task'



# This file is automatically generated during the generation of setup.py
# Copyright 2023-2026, Open XAI PnP
author = "Team SAIL"
author_email = "open.xai@gmail.com"
version = "0.1.3"
full_version = "0.1.3"
git_version = '78e93c7eecf360e21ffaf95a930d8790633be820'
release = True
homepage = "https://github.com/OpenXAIProject/pnpxai"



from typing import Callable, Optional, List, Tuple, Union, Sequence

from torch import Tensor
from torch.nn.modules import Module
from captum.attr import InputXGradient as CaptumGradientXInput
from captum.attr import LayerGradientXActivation as CaptumLayerGradientXInput

from pnpxai.core.detector.types import Linear, Convolution, LSTM, RNN, Attention
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils import captum_wrap_model_input


class GradientXInput(Explainer):
    """
    Grad X Input explainer.

    Supported Modules: `Linear`, `Convolution`, `LSTM`, `RNN`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        layer (Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]): The target module to be explained
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Avanti Shrikumar, Peyton Greenside, Anna Shcherbina, Anshul Kundaje. Not Just a Black Box: Learning Important Features Through Propagating Activation Differences.
    """

    SUPPORTED_MODULES = [Linear, Convolution, LSTM, RNN, Attention]

    def __init__(
        self,
        model: Module,
        layer: Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]] = None,
        forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
        additional_forward_arg_extractor: Optional[Callable[[Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
    ) -> None:
        super().__init__(model, forward_arg_extractor, additional_forward_arg_extractor)
        self.layer = layer


    @property
    def _layer_explainer(self) -> CaptumLayerGradientXInput:
        wrapped_model = captum_wrap_model_input(self.model)
        layers = [
            wrapped_model.input_maps[layer] if isinstance(layer, str)
            else layer for layer in self.layer
        ] if isinstance(self.layer, Sequence) else self.layer
        return CaptumLayerGradientXInput(
            forward_func=wrapped_model,
            layer=layers,
        )

    @property
    def _explainer(self) -> CaptumGradientXInput:
        return CaptumGradientXInput(forward_func=self.model)
    
    @property
    def explainer(self) -> Union[CaptumGradientXInput, CaptumLayerGradientXInput]:
        if self.layer is None:
            return self._explainer
        return self._layer_explainer

    def attribute(
        self,
        inputs: Union[Tensor, Tuple[Tensor]],
        targets: Tensor
    ) -> Union[Tensor, Tuple[Tensor]]:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor]]: The result of the explanation.
        """

        forward_args, additional_forward_args = self._extract_forward_args(inputs)
        attrs = self.explainer.attribute(
            inputs=forward_args,
            target=targets,
            additional_forward_args=additional_forward_args,
        )
        if isinstance(attrs, list):
            attrs = tuple(attrs)
        if isinstance(attrs, tuple) and len(attrs) == 1:
            attrs = attrs[0]
        return attrs


"""
Class for computing Integrated Gradients attributions for a given model and layer.

Args:
    model (torch.nn.Module): The PyTorch model to explain.
    layer (torch.nn.Module or List[str or torch.nn.Module]):
        The layer(s) for which to compute attributions. To target an input layer, a string of argument name
        is available.
    forward_arg_extractor (Callable[[Tuple[Tensor]], Tensor or Tuple[Tensor]] or None, optional):
        A function to extract arguments for each target layer from the tuple of inputs.
        Defaults to None.
    additional_forward_arg_extractor (Callable[[Tuple[Tensor]], Tensor or Tuple[Tensor]] or None, optional):
        A function to extract additional arguments not to be forwarded through the target layer.
        Defaults to None.
    baseline_fn (Callable or List[Callable] or None, optional):
        The function(s) to generate baseline of forward arguments. Must have same length as a tuple of
        forward arguments extracted by `forward_arg_extractor`. Defaults to None.
    n_step (int, optional): The number of steps for numerical approximation. Defaults to 20.

Raises:
    AssertionError: If the type of the `layer` or `baseline_fn` argument is not correct.

Example:
    For a given VQA model and dataset,
    ```
    # ./models.py

    class MyVQAModel(Module):
        ...

        def forward(img, qst, qst_len):
            x = self.vision_model(img)
            embedded = self.embedding(qst)
            y = self.question_model(embedded, qst_len)
            z = self.answer_model(x, y)
            return z
    
    # ./dataset.py
    class MyVQADataset(Dataset):
        ...

        def __getitem__(self, idx):
            ...
            return img, qst, qst_len
                
    ```

    Computes Integrated Gradients attributions for input image and embedded question.

    ```
    # model and data
    model = MyVQAModel().eval()
    dataloader = DataLoader(MyVQADataset(), batch_size=8, shuffle=False)
    imgs, qsts, qst_lens, answers = next(iter(dataloader))
    inputs = (imgs, qsts, qst_lens)
    outputs = model(*inputs)
    targets = outputs.argmax(1)

    # explainer
    layer_ig = LayerIntegratedGradients(
        model=model,
        layer=["img", model.embedding],
        forward_arg_extractor=lambda inputs: tuple(inputs[:2]),     # (imgs, qsts)
        additional_forward_arg_extractor=lambda inputs: inputs[-1], # qst_lens
        baseline_fn=[
            lambda imgs: torch.zeros_like(imgs),        # baseline function for images
            lambda qsts: torch.zeros_like(qsts).long(), # baseline function for questions
        ],
        n_step=20,
    )
    
    # attribute
    img_attrs, qst_attrs = layer_ig.attribute(inputs, targets)
    ```
"""


import _operator
from dataclasses import dataclass, asdict
from typing import Literal, List, Optional, Callable, Union

import torch
from torch.fx import Node, symbolic_trace, Tracer, GraphModule

SUPPORTED_FUNCTION_MODULES = {
    "torch": torch,
    "operator": _operator,
}

REPLACE_PREFIX = "_replaced_"


@dataclass
class NodeInfo:
    """
    Represents information about a node in a computation graph.

    Attributes:
    - opcode (Literal[str]): The operation code associated with the node.
    - name (str): The name of the node.
    - target (Union[Callable, str]): The target of the node, which could be a callable object or a string.
    """
    opcode: Literal[
        "placeholder", "get_attr", "call_function",
        "call_module", "call_method", "output",
    ]
    name: str
    target: Union[Callable, str]

    @classmethod
    def from_node(cls, n: Node):
        """
        Constructs a NodeInfo object from a given Node.

        Args:
        - n (Node): The node from which to construct the NodeInfo object.

        Returns:
        - NodeInfo: The NodeInfo object constructed from the given Node.
        """
        self = cls(
            opcode=n.op,
            name=n.name,
            target=n._pretty_print_target(n.target),
        )
        self._mode = "from_node"
        self._set_node(n.graph)
        return self

    @classmethod
    def from_module(cls, module: torch.nn.Module, **kwargs):
        """
        Constructs a NodeInfo object from a given torch.nn.Module.

        Args:
        - module (torch.nn.Module): The module from which to construct the NodeInfo object.
        - **kwargs: Additional keyword arguments.

        Returns:
        - NodeInfo: The NodeInfo object constructed from the given module.
        """
        self = cls(
            opcode="call_module",
            name=None,
            target=None,
        )
        self._mode = "from_module"
        self._set_operator(module)
        self._kwargs = kwargs
        return self

    @classmethod
    def from_function(cls, func: Callable, **kwargs):
        """
        Constructs a NodeInfo object from a given callable function.

        Args:
        - func (Callable): The function from which to construct the NodeInfo object.
        - **kwargs: Additional keyword arguments.

        Returns:
        - NodeInfo: The NodeInfo object constructed from the given function.
        """
        self = cls(
            opcode="call_function",
            name=None,
            target=None,
        )
        self._mode = "from_function"
        self._set_operator(func)
        self._kwargs = kwargs
        return self

    # set original node from the graph
    def _set_node(self, graph):
        assert self._mode == "from_node"
        # [GH] no `get_node` method in `torch.fx.Graph`. just find.
        for n in graph.nodes:
            if n.name == self.name:
                self._node = n

    def _set_operator(self, operator: Callable):
        assert self._mode != "from_node"
        self._valid_operator(operator)
        self._operator = operator
        return self

    # [TODO] validation for operator
    def _valid_operator(self, operator: Callable):
        pass

    def _get_operator(self, targets: List[str], root_module=None):
        operator = root_module if root_module else self._node.graph.owning_module
        for s in targets:
            operator = getattr(operator, s)
        return operator

    # cloning main attributions
    @property
    def meta(self):
        """
        Property to access the meta information of the node.

        Returns:
        - Any: The meta information of the node.
        """
        return self._node.meta

    @property
    def args(self):
        """
        Property to access the arguments of the node.

        Returns:
        - tuple: The arguments of the node.
        """
        return tuple([
            NodeInfo.from_node(a) if isinstance(a, Node) else a
            for a in self._node.args
        ])

    @property
    def kwargs(self):
        """
        Property to access the keyword arguments of the node.

        Returns:
        - dict: The keyword arguments of the node.
        """
        return {
            k: NodeInfo.from_node(v) if isinstance(v, Node) else v
            for k, v in self._node.kwargs.items()
        }

    @property
    def users(self):
        """
        Property to access the users of the node.

        Returns:
        - tuple: The users of the node.
        """
        return tuple([
            NodeInfo.from_node(u)
            for u in self._node.users.keys()
        ])

    @property
    def next(self):
        """
        Property to access the next node.

        Returns:
        - NodeInfo: The next node.
        """
        if self._node.next.op == "root":
            return
        return NodeInfo.from_node(self._node.next)

    @property
    def prev(self):
        """
        Property to access the previous node.

        Returns:
        - NodeInfo: The previous node.
        """
        if self._node.prev.op == "root":
            return
        return NodeInfo.from_node(self._node.prev)

    # additional properties for detection
    @property
    def operator(self) -> Optional[Callable]:
        """
        Property to access the operator.

        Returns:
        - Optional[Callable]: The operator.
        """
        if self._mode == "from_node":
            if self.opcode == "call_module":
                return self._get_operator(self.target.split("."))
            elif self.opcode == "call_function":
                targets = self.target.split(".")
                root_module = SUPPORTED_FUNCTION_MODULES.get(targets.pop(0))
                if root_module:
                    return self._get_operator(targets, root_module)
            return
        return self._operator

    @property
    def owning_module(self) -> Optional[str]:
        """
        Property to access the owning module.

        Returns:
        - Optional[str]: The owning module.
        """
        if self.opcode in ["call_module", "call_function"]:
            if self.meta.get("nn_module_stack"):
                nm = next(iter(self.meta["nn_module_stack"]))
                return nm
        return

    # convert data format
    def to_dict(self):
        """
        Converts the NodeInfo object to a dictionary.

        Returns:
        - dict: The dictionary representation of the NodeInfo object.
        """
        return {**asdict(self), "operator": self.operator}

    # [TODO] to_json for visualization
    def to_json_serializable(self):
        pass


class ModelArchitecture:
    """
    Represents the architecture of a model with methods for manipulating nodes.

    Attributes:
    - model: The model for which the architecture is defined.
    - tracer: A tracer to trace the model. By default, the  model is traced by `torch.fx.symbolic_trace`.
    """

    def __init__(self, model, tracer:Optional[Tracer]=None):
        self.model = model
        self.tracer = tracer

        self._traced_model = self._trace(model)
        self._replacing = False
    
    def _trace(self, model):
        if self.tracer is None:
            return symbolic_trace(model)
        graph = self.tracer.trace(model)
        name = (
            model.__class__.__name__ if isinstance(model, torch.nn.Module) else model.__name__
        )
        return GraphModule(self.tracer.root, graph, name)

    def list_nodes(self) -> List[NodeInfo]:
        """
        Lists all nodes in the model.

        Returns:
        - List[NodeInfo]: A list of NodeInfo objects representing the nodes in the model.
        """
        return [NodeInfo.from_node(n) for n in self._traced_model.graph.nodes]

    def get_node(self, name: str) -> NodeInfo:
        """
        Retrieves a node by its name.

        Args:
        - name (str): The name of the node to retrieve.

        Returns:
        - NodeInfo: The NodeInfo object corresponding to the specified node name.
        """
        for n in self._traced_model.graph.nodes:
            if n.name == name:
                return NodeInfo.from_node(n)
        return  # [TODO] Error for no result?

    def find_node(
        self,
        filter_func: Callable[[NodeInfo], bool],
        root: Optional[NodeInfo] = None,
        get_all: bool = False
    ) -> Union[NodeInfo, List[NodeInfo]]:
        """
        Finds a node based on a filtering function.

        Args:
            filter_func (Callable): The function used to filter nodes.
            root (Optional[NodeInfo]): The root node from which to start the search.
            get_all (bool): Whether to find all nodes matching the criteria.

        Returns:
            Union[NodeInfo, List[NodeInfo]]: The found node(s) or None if no node is found.
        """
        if root is None:
            # Take the first node
            root = NodeInfo.from_node(next(iter(self._traced_model.graph.nodes)))

        node = root
        nodes = []
        while node is not None:
            is_found = filter_func(node)
            if is_found:
                if not get_all:
                    return node
                nodes.append(node)
            node = node.next
        if len(nodes) == 0 and not get_all:
            return None
        return nodes

    def _validate_new_node(self, new_node: NodeInfo):
        if new_node.name is not None:
            exists = self.get_node(name=new_node.name)
            assert not exists, f"A node named {new_node.name} already exists."
        return True

    def _ensure_graph(self) -> None:
        self._traced_model.graph.lint()
        self._traced_model.recompile()

    def replace_node(self, node: NodeInfo, new_node: NodeInfo) -> NodeInfo:
        """
        Replaces a node in the model with a new node.

        Args:
        - node (NodeInfo): The node to replace.
        - new_node (NodeInfo): The new node to insert.

        Returns:
        - NodeInfo: The inserted node.
        """
        self._replacing = True
        self._validate_new_node(new_node)
        try:
            if new_node._mode == "from_module":
                new_node.name = f"{REPLACE_PREFIX}{node.name}"
            inserted = self.insert_node(new_node, base_node=node)
            self._traced_model.graph.erase_node(node._node)
            self._ensure_graph()
        finally:
            self._replacing = False
        return inserted

    def insert_node(self, new_node: NodeInfo, base_node: NodeInfo, before=False) -> NodeInfo:
        """
        Inserts a new node into the model.

        Args:
        - new_node (NodeInfo): The new node to insert.
        - base_node (NodeInfo): The node before which or after which to insert the new node.
        - before (bool): Whether to insert the new node before the base node.

        Returns:
        - NodeInfo: The inserted node.
        """
        self._validate_new_node(new_node)
        inserting = self._traced_model.graph.inserting_before if before else self._traced_model.graph.inserting_after
        if self._replacing:
            _inserted_args = base_node._node.args
            _inserted_kwargs = base_node._node.kwargs
            pass
        elif before:
            _inserted_args = tuple(
                arg for arg in base_node._node.args if isinstance(arg, Node))
            _inserted_kwargs = {
                kw: arg for kw, arg in base_node._node.kwargs.items() if isinstance(arg, Node)}
            _inserted_kwargs = {**_inserted_kwargs, **new_node._kwargs}
        else:
            _inserted_args = (base_node,)
            _inserted_kwargs = new_node._kwargs

        # [TODO] validation for new_node: if new_node._mode=="from_module" , new_node.name exists
        # insert
        with inserting(base_node._node):
            if new_node._mode == "from_module":
                self._traced_model.add_submodule(
                    new_node.name, new_node.operator)
                _inserted = self._traced_model.graph.call_module(
                    new_node.name,
                    _inserted_args,
                    _inserted_kwargs,
                )
            elif new_node._mode == "from_function":
                _inserted = self._traced_model.graph.call_function(
                    new_node.operator,
                    _inserted_args,
                    _inserted_kwargs,
                )

        if self._replacing or not before:
            base_node._node.replace_all_uses_with(_inserted)
            pass
        elif before:
            _not_inserted_args = [
                arg for arg in base_node._node.args if not isinstance(arg, Node)]
            base_node._node.args = tuple([_inserted] + _not_inserted_args)
            base_node._node.kwargs = {
                kw: arg for kw, arg in base_node._node.kwargs.items() if not isinstance(arg, Node)}
        self._ensure_graph()
        return NodeInfo.from_node(_inserted)

    def to_dict(self):
        """
        Converts the model architecture to a dictionary representation.

        Returns:
        - dict: A dictionary containing nodes and edges of the model architecture.
        """
        nodes = []
        edges = []
        for n in self.list_nodes():
            nodes.append(n.to_dict())
            edges += [{"source": n.name, "target": u.name} for u in n.users]
        return {"nodes": nodes, "edges": edges}



from typing import Callable, List, Dict, Any, Optional
from dataclasses import dataclass


@dataclass
class FactScoreOutput:
    atomic_facts: List[str]
    scores: List[Any]
    aggregated_score: Any


class FactScore:
    def __init__(
        self,
        atomic_fact_generator: Callable[[str], List[str]],
        knowledge_source: Callable[[str, str], List[Dict[str, str]]],
        scorer: Callable[[str, str, List[Dict[str, str]]], Any],
        aggregate_fn: Optional[Callable[[List[Any]], Any]]=None,
    ) -> None:
        self.atomic_fact_generator = atomic_fact_generator
        self.knowledge_source = knowledge_source
        self.scorer = scorer
        self.aggregate_fn = aggregate_fn

    def evaluate(self, topic: str, generation: str) -> FactScoreOutput:
        scores =[]
        atomic_facts = self.atomic_fact_generator(generation)
        for atomic_fact in atomic_facts:
            knowledges = self.knowledge_source(topic, atomic_fact)
            score = self.scorer(topic, atomic_fact, knowledges)
            scores.append(score)
        aggregated = self.aggregate_fn(scores) if self.aggregate_fn is not None else None
        return FactScoreOutput(
            atomic_facts=atomic_facts,
            scores=scores,
            aggregated_score=aggregated,
        )



from collections import OrderedDict
from typing import Sequence, Tuple

from torch import Tensor
import torch
from torch import nn
from torch.nn import functional as F

from pnpxai.explainers.rap.rules.functions import Transpose, Unsqueeze, View, MatMul, SoftMax
from pnpxai.explainers.rap.rules.base import RelProp, RelPropSeparate, RelPropSimple, _TensorOrTensors, safe_divide
from pnpxai.utils import linear_from_params


def parse_arg_by_name(name, idx, args=None, kwargs=None, default=None):
    args = args or []
    kwargs = kwargs or {}
    return kwargs.get(name, args[idx] if len(args) > idx else default)


class Dropout(RelProp):
    pass


class MaxPool2d(RelPropSimple):
    pass


class MaxPool1d(RelPropSimple):
    pass


class AdaptiveAvgPool2d(RelPropSimple):
    pass


class AdaptiveAvgPool1d(RelPropSimple):
    pass


class AvgPool1d(RelPropSimple):
    pass


class AvgPool2d(RelPropSimple):
    pass


class BatchNorm1d(RelProp):
    pass


class BatchNorm2d(RelProp):
    pass


class LayerNorm(RelProp):
    pass


class Linear(RelPropSeparate):
    @property
    def agg_dims(self):
        return -1

    def forward(self, weights: Tensor, inputs: Tensor) -> Tensor:
        return F.linear(weights, inputs)

    def first_prop(self, pd, px, nx, pw, nw):
        Rpp = self.forward(px, pw) * pd
        Rpn = self.forward(px, nw) * pd
        Rnp = self.forward(nx, pw) * pd
        Rnn = self.forward(nx, nw) * pd
        Pos = (Rpp + Rnn).sum(dim=-1, keepdim=True)
        Neg = (Rpn + Rnp).sum(dim=-1, keepdim=True)

        Z1 = self.forward(px, pw)
        Z2 = self.forward(px, nw)
        Z3 = self.forward(nx, pw)
        Z4 = self.forward(nx, nw)

        S1 = safe_divide(Rpp, Z1)
        S2 = safe_divide(Rpn, Z2)
        S3 = safe_divide(Rnp, Z3)
        S4 = safe_divide(Rnn, Z4)
        C1 = px * self.gradprop(Z1, px, S1)[0]
        C2 = px * self.gradprop(Z2, px, S2)[0]
        C3 = nx * self.gradprop(Z3, nx, S3)[0]
        C4 = nx * self.gradprop(Z4, nx, S4)[0]
        bp = self.module.bias * pd * safe_divide(Pos, Pos + Neg)
        bn = self.module.bias * pd * safe_divide(Neg, Pos + Neg)
        Sb1 = safe_divide(bp, Z1)
        Sb2 = safe_divide(bn, Z2)
        Cb1 = px * self.gradprop(Z1, px, Sb1)[0]
        Cb2 = px * self.gradprop(Z2, px, Sb2)[0]
        return C1 + C4 + Cb1 + C2 + C3 + Cb2

    def redistribute(self, Rp_tmp):
        Rp = torch.clamp(Rp_tmp, min=0)
        Rn = torch.clamp(Rp_tmp, max=0)
        R_tot = (Rp - Rn).sum(dim=-1, keepdim=True)
        Rp_tmp3 = safe_divide(Rp, R_tot) * \
            (Rp + Rn).sum(dim=-1, keepdim=True)
        Rn_tmp3 = -safe_divide(Rn, R_tot) * \
            (Rp + Rn).sum(dim=-1, keepdim=True)
        return Rp_tmp3 + Rn_tmp3

    def relprop(self, rel: _TensorOrTensors, inputs: _TensorOrTensors, outputs: _TensorOrTensors, args=None, kwargs=None):
        pos_w = torch.clamp(self.module.weight, min=0)
        neg_w = torch.clamp(self.module.weight, max=0)
        pos_x = torch.clamp(inputs, min=0)
        neg_x = torch.clamp(inputs, max=0)

        if torch.is_tensor(rel) and rel.max() == 1:  # first propagation
            rel = self.first_prop(rel, pos_x, neg_x, pos_w, neg_w)
            return self.redistribute(rel)

        return self.backward(rel, pos_x, neg_x, pos_w, neg_w)


class Conv2d(RelPropSeparate):
    @property
    def agg_dims(self):
        return list(range(1, 4))

    def forward(self, inputs: Tensor, weights: Tensor) -> Tensor:
        return torch.conv2d(inputs, weights, stride=self.module.stride, padding=self.module.padding)

    def gradprop_trans(self, inputs, outputs, DY, weight):
        padding = self.module.padding
        stride = self.module.stride
        kernel_size = self.module.kernel_size

        output_padding = inputs.size(2) - (
            (outputs.size(2) - 1) * stride[0] - 2 * padding[0] + kernel_size[0]
        )
        return F.conv_transpose2d(DY, weight, stride=stride, padding=padding, output_padding=output_padding)

    def final_backward(self, rel: Tensor, inputs: Tensor, outputs: Tensor, pos_w: Tensor, neg_w: Tensor):
        agg_dims = tuple(range(1, inputs.ndim))
        template = torch.zeros_like(inputs)
        low = template + torch.amin(inputs, dim=agg_dims, keepdim=True)
        high = template + torch.amax(inputs, dim=agg_dims, keepdim=True)

        norm_out = self.forward(inputs, self.module.weight) \
            - self.forward(low, pos_w) \
            - self.forward(high, neg_w)

        scale = safe_divide(rel, norm_out)

        rel = inputs * self.gradprop_trans(inputs, outputs, scale, self.module.weight) \
            - low * self.gradprop_trans(inputs, outputs, scale, pos_w) \
            - high * self.gradprop_trans(inputs, outputs, scale, neg_w)

        return rel

    def relprop(self, rel, inputs: _TensorOrTensors, outputs: _TensorOrTensors, args=None, kwargs=None):
        pos_w = torch.clamp(self.module.weight, min=0)
        neg_w = torch.clamp(self.module.weight, max=0)
        pos_x = torch.clamp(inputs, min=0)
        neg_x = torch.clamp(inputs, max=0)

        if inputs.shape[1] == 3:
            return self.final_backward(rel, inputs, outputs, pos_w, neg_w)

        return self.backward(rel, pos_x, neg_x, pos_w, neg_w)


class Conv1d(RelPropSeparate):
    @property
    def agg_dims(self):
        return list(range(1, 3))

    def forward(self, inputs: Tensor, weights: Tensor) -> Tensor:
        return torch.conv1d(inputs, weights, stride=self.module.stride, padding=self.module.padding)

    def relprop(self, rel, inputs: _TensorOrTensors, outputs: _TensorOrTensors, args=None, kwargs=None):
        pos_w = torch.clamp(self.module.weight, min=0)
        neg_w = torch.clamp(self.module.weight, max=0)
        pos_x = torch.clamp(inputs, min=0)
        neg_x = torch.clamp(inputs, max=0)

        return self.backward(rel, pos_x, neg_x, pos_w, neg_w)


class MultiHeadAttention(RelPropSimple):
    def __init__(self, module: nn.Module):
        self.module: nn.MultiheadAttention
        super(MultiHeadAttention, self).__init__(module)

    def _get_args(self, args, kwargs):
        return (
            parse_arg_by_name('key_padding_mask', 3, args, kwargs),
            parse_arg_by_name('need_weights', 4, args, kwargs, True),
            parse_arg_by_name('attn_mask', 5, args, kwargs),
            parse_arg_by_name('average_attn_weights', 6, args, kwargs, True),
            parse_arg_by_name('is_causal', 7, args, kwargs, False),
        )

    def _preprocess(self, rules: list, query: Tensor, key: Tensor, value: Tensor, key_padding_mask=None,  need_weights=True, attn_mask=None, is_causal=False) -> Tuple[Tuple[Tensor, Tensor, Tensor, Tensor], list]:
        is_batched = F._mha_shape_check(
            query, key, value, key_padding_mask, attn_mask, self.module.num_heads
        )
        if self.module.batch_first and is_batched:
            rules.append(lambda *rels: [
                Transpose().relprop(rel, x, None, (query, 1, 0), {})
                for rel, x in zip(rels, (query, key, value))
            ])
            query, key, value = (
                x.transpose(1, 0) for x in (query, key, value)
            )

        if not is_batched:
            rules.append(lambda *rels: [
                Unsqueeze().relprop(rel, x, None, (x, 1))
                for rel, x in zip(rels, (query, key, value))
            ])
            query, key, value = (
                x.unsqueeze(1) for x in (query, key, value)
            )
            if key_padding_mask is not None:
                key_padding_mask = key_padding_mask.unsqueeze(0)

        key_padding_mask = F._canonical_mask(
            mask=key_padding_mask,
            mask_name="key_padding_mask",
            other_type=F._none_or_dtype(attn_mask),
            other_name="attn_mask",
            target_type=query.dtype
        )

        if is_causal and key_padding_mask is None and not need_weights:
            attn_mask = None
        else:
            attn_mask = F._canonical_mask(
                mask=attn_mask,
                mask_name="attn_mask",
                other_type=None,
                other_name="",
                target_type=query.dtype,
                check_other=False,
            )

            if key_padding_mask is not None:
                is_causal = False

        return (query, key, value, attn_mask), rules

    def _qkv_proj(self, rules: list,  query: Tensor, key: Tensor, value: Tensor) -> Tuple[Tuple[Tensor, Tensor, Tensor], list]:
        batch_size = query.shape[1]

        if self.module._qkv_same_embed_dim:
            indices = [(i+1) * self.module.kdim for i in range(3)]
            w_q = self.module.in_proj_weight[:indices[0]]
            w_k = self.module.in_proj_weight[indices[0]:indices[1]]
            w_v = self.module.in_proj_weight[indices[1]:indices[2]]
        else:
            w_q, w_k, w_v = self.module.q_proj_weight, self.module.k_proj_weight, self.module.v_proj_weight

        if self.module.in_proj_bias is not None:
            b_q, b_k, b_v = self.module.in_proj_bias.chunk(3)
        else:
            b_q = b_k = b_v = None

        q = F.linear(query, w_q, b_q)
        k = F.linear(key, w_k, b_k)
        v = F.linear(value, w_v, b_v)

        rules.append(lambda *rels: [
            Linear(linear_from_params(w_q, b_q))
            .relprop(rel, x, None)
            for rel, x in zip(rels, (query, key, value))
        ])

        if self.module.bias_k is not None:
            k = torch.cat([k, self.module.bias_k.repeat(1, batch_size, 1)])
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))

        return (q, k, v), rules

    def _view_with_head(self, rules: list, query: Tensor, key: Tensor, value: Tensor) -> Tuple[Tuple[Tensor, Tensor, Tensor], list]:
        batch_size = value.shape[1]
        num_heads = self.module.num_heads
        head_dim = self.module.head_dim

        def _get_view_shape(x: Tensor):
            return x.shape[0], batch_size * num_heads, head_dim

        query_t, key_t, value_t = (
            x.view(_get_view_shape(x)).transpose(0, 1)
            for x in (query, key, value)
        )

        rules.append(lambda *rels: [
            View().relprop(rel, x, None, (x, *_get_view_shape(x)))
            for rel, x in zip(rels, (query, key, value))
        ])
        rules.append(lambda *rels: [
            Transpose().relprop(rel, None, None, (None, 0, 1), {})
            for rel, x in zip(rels, (query, key, value))
        ])

        return (query_t, key_t, value_t), rules

    def _attention(self, rules: list, query: Tensor, key: Tensor, key_padding_mask=None, attn_mask=None) -> Tuple[Tensor, list]:
        batch_size = query.shape[1]
        num_heads = self.module.num_heads
        head_dim = self.module.head_dim

        if self.module.add_zero_attn:
            zero_attn_shape = (batch_size * num_heads, 1, head_dim)
            key_zeros = torch.zeros(
                zero_attn_shape, dtype=key.dtype, device=key.device)
            key = torch.cat([key, key_zeros], dim=1)
            if attn_mask is not None:
                attn_mask = F.pad(attn_mask, (0, 1))
            if key_padding_mask is not None:
                key_padding_mask = F.pad(key_padding_mask, (0, 1))

        if key_padding_mask is not None:
            src_len = key.shape(1)
            assert key_padding_mask.shape == (batch_size, src_len), \
                f"expecting key_padding_mask shape of {(batch_size, src_len)}, but got {key_padding_mask.shape}"
            key_padding_mask = key_padding_mask.view(batch_size, 1, 1, src_len) \
                .expand(-1, num_heads, -1, -1)\
                .reshape(batch_size * num_heads, 1, src_len)
            if attn_mask is None:
                attn_mask = key_padding_mask
            else:
                attn_mask = attn_mask + key_padding_mask

        E = query.shape[-1]
        key_t = key.transpose(-2, -1)
        attn_out = torch.bmm(query, key_t) / (E**.5)

        rules.append(lambda q_rel, k_rel, v_rel: [
            q_rel, Transpose().relprop(k_rel, key, key_t, (key, -2, -1), {}), v_rel
        ])
        rules.append(lambda attn_rel, v_rel: [
            *MatMul().relprop(attn_rel, (query, key_t), attn_out), v_rel
        ])
        if attn_mask is not None:
            attn_out = attn_mask + attn_out

        attn_out_weighted = F.softmax(attn_out, dim=-1)
        rules.append(lambda attn_rel, v_rel: [
            SoftMax().relprop(attn_rel, attn_out, attn_out_weighted, [-1], {}),
            v_rel
        ])

        return attn_out_weighted, rules

    def _attn_v_mul(self, rules: list, attn_weights: Tensor, value: Tensor) -> Tuple[Tensor, list]:
        out = torch.bmm(attn_weights, value)
        rules.append(lambda out_rel: [
            *MatMul().relprop(out_rel, (attn_weights, value), out),
        ])

        return out, rules

    def _out_proj(self, rules: list, out: Tensor, source_shape) -> Tuple[Tensor, list]:
        orig_out = out.clone()
        rules.append(lambda out_rel: [
            View().relprop(out_rel, orig_out, None)
        ])
        rules.append(lambda out_rel: [
            Transpose().relprop(out_rel, None, None, (None, 0, 1), {})
        ])

        out = out.view(source_shape).transpose(0, 1)

        out_lin = self.module.out_proj
        out_proj = out_lin(out)

        rules.append(lambda out_rel: (
            Linear(out_lin).relprop(out_rel, out, out_proj),
        ))

        return out_proj, rules

    def _match_rels_with_inputs(self, rels: Sequence[Tensor], inputs: Sequence[Tensor]):
        rel_map = OrderedDict()
        for rel, datum in zip(rels, inputs):
            idx = id(datum)
            if idx not in rel_map:
                rel_map[idx] = 0
            rel_map[idx] += rel

        return list(rel_map.values())

    def relprop(self, rel: Tensor, inputs: Tensor, outputs: Tensor, args=None, kwargs=None) -> _TensorOrTensors:
        if torch.is_tensor(inputs):
            inputs = (inputs, inputs, inputs)
        query, key, value = inputs
        key_padding_mask, need_weights, attn_mask, average_attn_weights, is_causal = self._get_args(
            args, kwargs
        )

        rules = []

        (query, key, value, attn_mask), rules = self._preprocess(
            rules, query, key, value, key_padding_mask, need_weights, attn_mask, is_causal
        )
        source_shape = query.shape

        (query, key, value), rules = self._qkv_proj(rules, query, key, value)

        (query, key, value), rules = self._view_with_head(rules, query, key, value)

        (attn_weights), rules = self._attention(
            rules, query, key, key_padding_mask, attn_mask
        )

        (out), rules = self._attn_v_mul(rules, attn_weights, value)

        (out), rules = self._out_proj(rules, out, source_shape)

        rels = (rel,)
        for rule in reversed(rules):
            rels = rule(*rels)

        rels = self._match_rels_with_inputs(inputs, rels)
        if len(rels) == 1:
            return rels[0]
        
        return rels


from typing import Callable, Optional, Tuple, Union, Sequence, Dict

from torch import Tensor
from torch.nn.modules import Module
from captum.attr import IntegratedGradients as CaptumIntegratedGradients
from captum.attr import LayerIntegratedGradients as CaptumLayerIntegratedGradients

from pnpxai.core.detector.types import Linear, Convolution, Attention
from pnpxai.utils import format_into_tuple, format_out_tuple_if_single
from pnpxai.explainers.utils.baselines import BaselineMethodOrFunction, BaselineFunction
from pnpxai.explainers.base import Explainer
from pnpxai.explainers.utils import captum_wrap_model_input


class IntegratedGradients(Explainer):
    """
    IntegratedGradients explainer.

    Supported Modules: `Linear`, `Convolution`, `Attention`

    Parameters:
        model (Module): The PyTorch model for which attribution is to be computed.
        baseline_fn (Union[BaselineMethodOrFunction, Tuple[BaselineMethodOrFunction]]): The baseline function, accepting the attribution input, and returning the baseline accordingly.
        n_steps (int): The Number of steps the algorithm makes
        layer (Optional[Union[Union[str, Module], Sequence[Union[str, Module]]]]): The target module to be explained
        n_classes (Optional[int]): Number of classes
        forward_arg_extractor: A function that extracts forward arguments from the input batch(s) where the attribution scores are assigned.
        additional_forward_arg_extractor: A secondary function that extract additional forward arguments from the input batch(s).
        **kwargs: Keyword arguments that are forwarded to the base implementation of the Explainer

    Reference:
        Mukund Sundararajan, Ankur Taly, Qiqi Yan. Axiomatic Attribution for Deep Networks.
    """

    SUPPORTED_MODULES = [Linear, Convolution, Attention]

    def __init__(
        self,
        model: Module,
        n_steps: int = 20,
        baseline_fn: Union[BaselineMethodOrFunction,
                           Tuple[BaselineMethodOrFunction]] = 'zeros',
        layer: Optional[Callable[[Tuple[Tensor]],
                                 Union[Tensor, Tuple[Tensor]]]] = None,
        forward_arg_extractor: Optional[Callable[[
            Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
        additional_forward_arg_extractor: Optional[Callable[[
            Tuple[Tensor]], Union[Tensor, Tuple[Tensor]]]] = None,
    ) -> None:
        super().__init__(model, forward_arg_extractor, additional_forward_arg_extractor)
        self.layer = layer
        self.n_steps = n_steps
        self.baseline_fn = baseline_fn

    @property
    def _layer_explainer(self) -> CaptumLayerIntegratedGradients:
        wrapped_model = captum_wrap_model_input(self.model)
        layers = [
            wrapped_model.input_maps[layer] if isinstance(layer, str)
            else layer for layer in self.layer
        ] if isinstance(self.layer, Sequence) else self.layer
        return CaptumLayerIntegratedGradients(
            forward_func=wrapped_model,
            layer=layers,
        )

    @property
    def _explainer(self) -> CaptumIntegratedGradients:
        return CaptumIntegratedGradients(forward_func=self.model)

    @property
    def explainer(self) -> Union[CaptumIntegratedGradients, CaptumLayerIntegratedGradients]:
        if self.layer is None:
            return self._explainer
        return self._layer_explainer

    def attribute(
        self,
        inputs: Union[Tensor, Tuple[Tensor]],
        targets: Tensor
    ) -> Union[Tensor, Tuple[Tensor]]:
        """
        Computes attributions for the given inputs and targets.

        Args:
            inputs (torch.Tensor): The input data.
            targets (torch.Tensor): The target labels for the inputs.

        Returns:
            Union[torch.Tensor, Tuple[torch.Tensor]]: The result of the explanation.
        """
        forward_args, additional_forward_args = self._extract_forward_args(
            inputs)
        forward_args = format_into_tuple(forward_args)
        baselines = format_into_tuple(self._get_baselines(forward_args))
        attrs = self.explainer.attribute(
            inputs=forward_args,
            baselines=baselines,
            target=targets,
            additional_forward_args=additional_forward_args,
            n_steps=self.n_steps,
        )
        if isinstance(attrs, tuple):
            attrs = format_out_tuple_if_single(attrs)
        return attrs

    def get_tunables(self) -> Dict[str, Tuple[type, dict]]:
        """
        Provides Tunable parameters for the optimizer

        Tunable parameters:
            `noise_level` (float): Value can be selected in the range of `range(10, 100, 10)`

            `baseline_fn` (callable): BaselineFunction selects suitable values in accordance with the modality
        """
        return {
            'n_steps': (int, {'low': 10, 'high': 100, 'step': 10}),
            'baseline_fn': (BaselineFunction, {}),
        }
